import numpy as np
from spglm.family import Gaussian, Binomial, Poisson
from spglm.glm import GLM
from .diagnostics import get_AICc


def summaryModel(self):
    summary = '=' * 75 + '\n'
    summary += "%-54s %20s\n" % ('Model type', self.family.__class__.__name__)
    summary += "%-60s %14d\n" % ('Number of observations:', self.n)
    summary += "%-60s %14d\n\n" % ('Number of covariates:', self.k)
    return summary


def summaryGLM(self):

    XNames = ["X" + str(i) for i in range(self.k)]
    glm_rslt = GLM(self.model.y, self.model.X, constant=False,
                   family=self.family).fit()

    summary = "%s\n" % ('Global Regression Results')
    summary += '-' * 75 + '\n'

    if isinstance(self.family, Gaussian):
        summary += "%-62s %12.3f\n" % ('Residual sum of squares:',
                                       glm_rslt.deviance)
        summary += "%-62s %12.3f\n" % ('Log-likelihood:', glm_rslt.llf)
        summary += "%-62s %12.3f\n" % ('AIC:', glm_rslt.aic)
        summary += "%-62s %12.3f\n" % ('AICc:', get_AICc(glm_rslt))
        summary += "%-62s %12.3f\n" % ('BIC:', glm_rslt.bic)
        summary += "%-62s %12.3f\n" % ('R2:', glm_rslt.D2)
        summary += "%-62s %12.3f\n\n" % ('Adj. R2:', glm_rslt.adj_D2)
    else:
        summary += "%-62s %12.3f\n" % ('Deviance:', glm_rslt.deviance)
        summary += "%-62s %12.3f\n" % ('Log-likelihood:', glm_rslt.llf)
        summary += "%-62s %12.3f\n" % ('AIC:', glm_rslt.aic)
        summary += "%-62s %12.3f\n" % ('AICc:', get_AICc(glm_rslt))
        summary += "%-62s %12.3f\n" % ('BIC:', glm_rslt.bic)
        summary += "%-62s %12.3f\n" % ('Percent deviance explained:',
                                       glm_rslt.D2)
        summary += "%-62s %12.3f\n\n" % ('Adj. percent deviance explained:',
                                         glm_rslt.adj_D2)

    summary += "%-31s %10s %10s %10s %10s\n" % ('Variable', 'Est.', 'SE',
                                                't(Est/SE)', 'p-value')
    summary += "%-31s %10s %10s %10s %10s\n" % ('-' * 31, '-' * 10, '-' * 10,
                                                '-' * 10, '-' * 10)
    for i in range(self.k):
        summary += "%-31s %10.3f %10.3f %10.3f %10.3f\n" % (
            XNames[i], glm_rslt.params[i], glm_rslt.bse[i],
            glm_rslt.tvalues[i], glm_rslt.pvalues[i])
    summary += "\n"
    return summary


def summaryGWR(self):
    XNames = ["X" + str(i) for i in range(self.k)]

    summary = "%s\n" % ('Geographically Weighted Regression (GWR) Results')
    summary += '-' * 75 + '\n'

    if self.model.fixed:
        summary += "%-50s %20s\n" % ('Spatial kernel:',
                                     'Fixed ' + self.model.kernel)
    else:
        summary += "%-54s %20s\n" % ('Spatial kernel:',
                                     'Adaptive ' + self.model.kernel)

    summary += "%-62s %12.3f\n" % ('Bandwidth used:', self.model.bw)

    summary += "\n%s\n" % ('Diagnostic information')
    summary += '-' * 75 + '\n'

    if isinstance(self.family, Gaussian):
        summary += "%-62s %12.3f\n" % ('Residual sum of squares:',
                                       self.resid_ss)
        summary += "%-62s %12.3f\n" % (
            'Effective number of parameters (trace(S)):', self.tr_S)
        summary += "%-62s %12.3f\n" % ('Degree of freedom (n - trace(S)):',
                                       self.df_model)
        summary += "%-62s %12.3f\n" % ('Sigma estimate:', np.sqrt(self.sigma2))
        summary += "%-62s %12.3f\n" % ('Log-likelihood:', self.llf)
        summary += "%-62s %12.3f\n" % ('AIC:', self.aic)
        summary += "%-62s %12.3f\n" % ('AICc:', self.aicc)
        summary += "%-62s %12.3f\n" % ('BIC:', self.bic)
        summary += "%-62s %12.3f\n" % ('R2:', self.R2)
        summary += "%-62s %12.3f\n" % ('Adjusted R2:', self.adj_R2)

    else:
        summary += "%-62s %12.3f\n" % (
            'Effective number of parameters (trace(S)):', self.tr_S)
        summary += "%-62s %12.3f\n" % ('Degree of freedom (n - trace(S)):',
                                       self.df_model)
        summary += "%-62s %12.3f\n" % ('Log-likelihood:', self.llf)
        summary += "%-62s %12.3f\n" % ('AIC:', self.aic)
        summary += "%-62s %12.3f\n" % ('AICc:', self.aicc)
        summary += "%-62s %12.3f\n" % ('BIC:', self.bic)
        summary += "%-60s %12.3f\n" % ('Percent deviance explained:', self.D2)
        summary += "%-60s %12.3f\n" % ('Adjusted percent deviance explained:',
                                       self.adj_D2)

    summary += "%-62s %12.3f\n" % ('Adj. alpha (95%):', self.adj_alpha[1])
    summary += "%-62s %12.3f\n" % ('Adj. critical t value (95%):',
                                   self.critical_tval(self.adj_alpha[1]))

    summary += "\n%s\n" % ('Summary Statistics For GWR Parameter Estimates')
    summary += '-' * 75 + '\n'
    summary += "%-20s %10s %10s %10s %10s %10s\n" % ('Variable', 'Mean', 'STD',
                                                     'Min', 'Median', 'Max')
    summary += "%-20s %10s %10s %10s %10s %10s\n" % (
        '-' * 20, '-' * 10, '-' * 10, '-' * 10, '-' * 10, '-' * 10)
    for i in range(self.k):
        summary += "%-20s %10.3f %10.3f %10.3f %10.3f %10.3f\n" % (
            XNames[i], np.mean(self.params[:, i]), np.std(self.params[:, i]),
            np.min(self.params[:, i]), np.median(self.params[:, i]),
            np.max(self.params[:, i]))

    summary += '=' * 75 + '\n'

    return summary


def summaryMGWR(self):

    XNames = ["X" + str(i) for i in range(self.k)]

    summary = ''
    summary += "%s\n" % (
        'Multi-Scale Geographically Weighted Regression (MGWR) Results')
    summary += '-' * 75 + '\n'

    if self.model.fixed:
        summary += "%-50s %20s\n" % ('Spatial kernel:',
                                     'Fixed ' + self.model.kernel)
    else:
        summary += "%-54s %20s\n" % ('Spatial kernel:',
                                     'Adaptive ' + self.model.kernel)

    summary += "%-54s %20s\n" % ('Criterion for optimal bandwidth:',
                                 self.model.selector.criterion)

    if self.model.selector.rss_score:
        summary += "%-54s %20s\n" % ('Score of Change (SOC) type:', 'RSS')
    else:
        summary += "%-54s %20s\n" % ('Score of Change (SOC) type:',
                                     'Smoothing f')

    summary += "%-54s %20s\n\n" % ('Termination criterion for MGWR:',
                                   self.model.selector.tol_multi)

    summary += "%s\n" % ('MGWR bandwidths')
    summary += '-' * 75 + '\n'
    summary += "%-15s %14s %10s %16s %16s\n" % (
        'Variable', 'Bandwidth', 'ENP_j', 'Adj t-val(95%)', 'Adj alpha(95%)')
    for j in range(self.k):
        summary += "%-14s %15.3f %10.3f %16.3f %16.3f\n" % (
            XNames[j], self.model.bws[j], self.ENP_j[j],
            self.critical_tval()[j], self.adj_alpha_j[j, 1])

    summary += "\n%s\n" % ('Diagnostic information')
    summary += '-' * 75 + '\n'

    summary += "%-62s %12.3f\n" % ('Residual sum of squares:', self.resid_ss)
    summary += "%-62s %12.3f\n" % (
        'Effective number of parameters (trace(S)):', self.tr_S)
    summary += "%-62s %12.3f\n" % ('Degree of freedom (n - trace(S)):',
                                   self.df_model)

    summary += "%-62s %12.3f\n" % ('Sigma estimate:', np.sqrt(self.sigma2))
    summary += "%-62s %12.3f\n" % ('Log-likelihood:', self.llf)
    summary += "%-62s %12.3f\n" % ('AIC:', self.aic)
    summary += "%-62s %12.3f\n" % ('AICc:', self.aicc)
    summary += "%-62s %12.3f\n" % ('BIC:', self.bic)
    summary += "%-62s %12.3f\n" % ('R2', self.R2)
    summary += "%-62s %12.3f\n" % ('Adjusted R2', self.adj_R2)

    summary += "\n%s\n" % ('Summary Statistics For MGWR Parameter Estimates')
    summary += '-' * 75 + '\n'
    summary += "%-20s %10s %10s %10s %10s %10s\n" % ('Variable', 'Mean', 'STD',
                                                     'Min', 'Median', 'Max')
    summary += "%-20s %10s %10s %10s %10s %10s\n" % (
        '-' * 20, '-' * 10, '-' * 10, '-' * 10, '-' * 10, '-' * 10)
    for i in range(self.k):
        summary += "%-20s %10.3f %10.3f %10.3f %10.3f %10.3f\n" % (
            XNames[i], np.mean(self.params[:, i]), np.std(self.params[:, i]),
            np.min(self.params[:, i]), np.median(self.params[:, i]),
            np.max(self.params[:, i]))

    summary += '=' * 75 + '\n'
    return summary



"""
GWR is tested against results from GWR4
"""

import os
import libpysal as ps
from libpysal import io
import numpy as np
import unittest
import pandas
from types import SimpleNamespace
from ..gwr import GWR, MGWR, MGWRResults
from ..sel_bw import Sel_BW
from ..diagnostics import get_AICc, get_AIC, get_BIC, get_CV
from spglm.family import Gaussian, Poisson, Binomial


class TestGWRGaussian(unittest.TestCase):
    def setUp(self):
        data_path = ps.examples.get_path("GData_utm.csv")
        data = io.open(data_path)
        self.coords = list(zip(data.by_col('X'), data.by_col('Y')))
        self.y = np.array(data.by_col('PctBach')).reshape((-1, 1))
        rural = np.array(data.by_col('PctRural')).reshape((-1, 1))
        pov = np.array(data.by_col('PctPov')).reshape((-1, 1))
        black = np.array(data.by_col('PctBlack')).reshape((-1, 1))
        fb = np.array(data.by_col('PctFB')).reshape((-1, 1))
        self.X = np.hstack([rural, pov, black])
        self.mgwr_X = np.hstack([fb, black, rural])
        self.BS_F = io.open(ps.examples.get_path('georgia_BS_F_listwise.csv'))
        self.BS_NN = io.open(
            ps.examples.get_path('georgia_BS_NN_listwise.csv'))
        self.GS_F = io.open(ps.examples.get_path('georgia_GS_F_listwise.csv'))
        self.GS_NN = io.open(
            ps.examples.get_path('georgia_GS_NN_listwise.csv'))
        MGWR_path = os.path.join(
            os.path.dirname(__file__), 'georgia_mgwr_results.csv')
        self.MGWR = pandas.read_csv(MGWR_path)

    def test_BS_F(self):
        est_Int = self.BS_F.by_col(' est_Intercept')
        se_Int = self.BS_F.by_col(' se_Intercept')
        t_Int = self.BS_F.by_col(' t_Intercept')
        est_rural = self.BS_F.by_col(' est_PctRural')
        se_rural = self.BS_F.by_col(' se_PctRural')
        t_rural = self.BS_F.by_col(' t_PctRural')
        est_pov = self.BS_F.by_col(' est_PctPov')
        se_pov = self.BS_F.by_col(' se_PctPov')
        t_pov = self.BS_F.by_col(' t_PctPov')
        est_black = self.BS_F.by_col(' est_PctBlack')
        se_black = self.BS_F.by_col(' se_PctBlack')
        t_black = self.BS_F.by_col(' t_PctBlack')
        yhat = self.BS_F.by_col(' yhat')
        res = np.array(self.BS_F.by_col(' residual'))
        std_res = np.array(self.BS_F.by_col(' std_residual')).reshape((-1, 1))
        localR2 = np.array(self.BS_F.by_col(' localR2')).reshape((-1, 1))
        inf = np.array(self.BS_F.by_col(' influence')).reshape((-1, 1))
        cooksD = np.array(self.BS_F.by_col(' CooksD')).reshape((-1, 1))

        model = GWR(self.coords, self.y, self.X, bw=209267.689, fixed=True,
                    sigma2_v1=False)

        rslt = model.fit()
        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)
        CV = get_CV(rslt)

        self.assertAlmostEquals(np.floor(AICc), 894.0)
        self.assertAlmostEquals(np.floor(AIC), 890.0)
        self.assertAlmostEquals(np.floor(BIC), 944.0)
        self.assertAlmostEquals(np.round(CV, 2), 18.25)
        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-04)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-04)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-04)
        np.testing.assert_allclose(est_rural, rslt.params[:, 1], rtol=1e-04)
        np.testing.assert_allclose(se_rural, rslt.bse[:, 1], rtol=1e-04)
        np.testing.assert_allclose(t_rural, rslt.tvalues[:, 1], rtol=1e-04)
        np.testing.assert_allclose(est_pov, rslt.params[:, 2], rtol=1e-04)
        np.testing.assert_allclose(se_pov, rslt.bse[:, 2], rtol=1e-04)
        np.testing.assert_allclose(t_pov, rslt.tvalues[:, 2], rtol=1e-04)
        np.testing.assert_allclose(est_black, rslt.params[:, 3], rtol=1e-02)
        np.testing.assert_allclose(se_black, rslt.bse[:, 3], rtol=1e-02)
        np.testing.assert_allclose(t_black, rslt.tvalues[:, 3], rtol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-05)
        np.testing.assert_allclose(res, rslt.resid_response, rtol=1e-04)
        np.testing.assert_allclose(std_res, rslt.std_res, rtol=1e-04)
        np.testing.assert_allclose(localR2, rslt.localR2, rtol=1e-05)
        np.testing.assert_allclose(inf, rslt.influ, rtol=1e-04)
        np.testing.assert_allclose(cooksD, rslt.cooksD, rtol=1e-00)

    def test_BS_NN(self):
        est_Int = self.BS_NN.by_col(' est_Intercept')
        se_Int = self.BS_NN.by_col(' se_Intercept')
        t_Int = self.BS_NN.by_col(' t_Intercept')
        est_rural = self.BS_NN.by_col(' est_PctRural')
        se_rural = self.BS_NN.by_col(' se_PctRural')
        t_rural = self.BS_NN.by_col(' t_PctRural')
        est_pov = self.BS_NN.by_col(' est_PctPov')
        se_pov = self.BS_NN.by_col(' se_PctPov')
        t_pov = self.BS_NN.by_col(' t_PctPov')
        est_black = self.BS_NN.by_col(' est_PctBlack')
        se_black = self.BS_NN.by_col(' se_PctBlack')
        t_black = self.BS_NN.by_col(' t_PctBlack')
        yhat = self.BS_NN.by_col(' yhat')
        res = np.array(self.BS_NN.by_col(' residual'))
        std_res = np.array(self.BS_NN.by_col(' std_residual')).reshape((-1, 1))
        localR2 = np.array(self.BS_NN.by_col(' localR2')).reshape((-1, 1))
        inf = np.array(self.BS_NN.by_col(' influence')).reshape((-1, 1))
        cooksD = np.array(self.BS_NN.by_col(' CooksD')).reshape((-1, 1))
        local_corr = os.path.join(os.path.dirname(__file__), 'local_corr.csv')
        corr1 = np.array(io.open(local_corr))
        local_vif = os.path.join(os.path.dirname(__file__), 'local_vif.csv')
        vif1 = np.array(io.open(local_vif))
        local_cn = os.path.join(os.path.dirname(__file__), 'local_cn.csv')
        cn1 = np.array(io.open(local_cn))
        local_vdp = os.path.join(os.path.dirname(__file__), 'local_vdp.csv')
        vdp1 = np.array(io.open(local_vdp), dtype=np.float64)
        spat_var_p_vals = [0., 0.0, 0.5, 0.2]

        model = GWR(self.coords, self.y, self.X, bw=90.000, fixed=False,
                    sigma2_v1=False)
        rslt = model.fit()

        adj_alpha = rslt.adj_alpha
        alpha = 0.01017489
        critical_t = rslt.critical_tval(alpha)
        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)
        CV = get_CV(rslt)
        corr2, vif2, cn2, vdp2 = rslt.local_collinearity()
        R2 = rslt.R2

        np.testing.assert_allclose(
            adj_alpha, np.array([0.02034978, 0.01017489, 0.0002035]),
            rtol=1e-04)
        self.assertAlmostEquals(critical_t, 2.6011011542649394)
        self.assertAlmostEquals(np.around(R2, 4), 0.5924)
        self.assertAlmostEquals(np.floor(AICc), 896.0)
        self.assertAlmostEquals(np.floor(AIC), 892.0)
        self.assertAlmostEquals(np.floor(BIC), 941.0)
        self.assertAlmostEquals(np.around(CV, 2), 19.19)
        np.testing.assert_allclose(corr1, corr2, rtol=1e-04)
        np.testing.assert_allclose(vif1, vif2, rtol=1e-04)
        np.testing.assert_allclose(cn1, cn2, rtol=1e-04)
        np.testing.assert_allclose(vdp1, vdp2, rtol=1e-04)

        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-04)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-04)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-04)
        np.testing.assert_allclose(est_rural, rslt.params[:, 1], rtol=1e-04)
        np.testing.assert_allclose(se_rural, rslt.bse[:, 1], rtol=1e-04)
        np.testing.assert_allclose(t_rural, rslt.tvalues[:, 1], rtol=1e-04)
        np.testing.assert_allclose(est_pov, rslt.params[:, 2], rtol=1e-04)
        np.testing.assert_allclose(se_pov, rslt.bse[:, 2], rtol=1e-04)
        np.testing.assert_allclose(t_pov, rslt.tvalues[:, 2], rtol=1e-04)
        np.testing.assert_allclose(est_black, rslt.params[:, 3], rtol=1e-02)
        np.testing.assert_allclose(se_black, rslt.bse[:, 3], rtol=1e-02)
        np.testing.assert_allclose(t_black, rslt.tvalues[:, 3], rtol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-05)
        np.testing.assert_allclose(res, rslt.resid_response, rtol=1e-04)
        np.testing.assert_allclose(std_res, rslt.std_res, rtol=1e-04)
        np.testing.assert_allclose(localR2, rslt.localR2, rtol=1e-05)
        np.testing.assert_allclose(inf, rslt.influ, rtol=1e-04)
        np.testing.assert_allclose(cooksD, rslt.cooksD, rtol=1e-00)

        sel = Sel_BW(self.coords, self.y, self.X)
        bw = sel.search()
        model = GWR(self.coords, self.y, self.X, bw)
        result = model.fit()

        p_vals = result.spatial_variability(sel, 10)
        np.testing.assert_allclose(spat_var_p_vals, p_vals, rtol=1e-04)

    def test_GS_F(self):
        est_Int = self.GS_F.by_col(' est_Intercept')
        se_Int = self.GS_F.by_col(' se_Intercept')
        t_Int = self.GS_F.by_col(' t_Intercept')
        est_rural = self.GS_F.by_col(' est_PctRural')
        se_rural = self.GS_F.by_col(' se_PctRural')
        t_rural = self.GS_F.by_col(' t_PctRural')
        est_pov = self.GS_F.by_col(' est_PctPov')
        se_pov = self.GS_F.by_col(' se_PctPov')
        t_pov = self.GS_F.by_col(' t_PctPov')
        est_black = self.GS_F.by_col(' est_PctBlack')
        se_black = self.GS_F.by_col(' se_PctBlack')
        t_black = self.GS_F.by_col(' t_PctBlack')
        yhat = self.GS_F.by_col(' yhat')
        res = np.array(self.GS_F.by_col(' residual'))
        std_res = np.array(self.GS_F.by_col(' std_residual')).reshape((-1, 1))
        localR2 = np.array(self.GS_F.by_col(' localR2')).reshape((-1, 1))
        inf = np.array(self.GS_F.by_col(' influence')).reshape((-1, 1))
        cooksD = np.array(self.GS_F.by_col(' CooksD')).reshape((-1, 1))

        model = GWR(self.coords, self.y, self.X, bw=87308.298,
                    kernel='gaussian', fixed=True, sigma2_v1=False)

        rslt = model.fit()
        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)
        CV = get_CV(rslt)

        self.assertAlmostEquals(np.floor(AICc), 895.0)
        self.assertAlmostEquals(np.floor(AIC), 890.0)
        self.assertAlmostEquals(np.floor(BIC), 943.0)
        self.assertAlmostEquals(np.around(CV, 2), 18.21)
        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-04)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-04)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-04)
        np.testing.assert_allclose(est_rural, rslt.params[:, 1], rtol=1e-04)
        np.testing.assert_allclose(se_rural, rslt.bse[:, 1], rtol=1e-04)
        np.testing.assert_allclose(t_rural, rslt.tvalues[:, 1], rtol=1e-04)
        np.testing.assert_allclose(est_pov, rslt.params[:, 2], rtol=1e-04)
        np.testing.assert_allclose(se_pov, rslt.bse[:, 2], rtol=1e-04)
        np.testing.assert_allclose(t_pov, rslt.tvalues[:, 2], rtol=1e-04)
        np.testing.assert_allclose(est_black, rslt.params[:, 3], rtol=1e-02)
        np.testing.assert_allclose(se_black, rslt.bse[:, 3], rtol=1e-02)
        np.testing.assert_allclose(t_black, rslt.tvalues[:, 3], rtol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-05)
        np.testing.assert_allclose(res, rslt.resid_response, rtol=1e-04)
        np.testing.assert_allclose(std_res, rslt.std_res, rtol=1e-04)
        np.testing.assert_allclose(localR2, rslt.localR2, rtol=1e-05)
        np.testing.assert_allclose(inf, rslt.influ, rtol=1e-04)
        np.testing.assert_allclose(cooksD, rslt.cooksD, rtol=1e-00)

    def test_GS_NN(self):
        est_Int = self.GS_NN.by_col(' est_Intercept')
        se_Int = self.GS_NN.by_col(' se_Intercept')
        t_Int = self.GS_NN.by_col(' t_Intercept')
        est_rural = self.GS_NN.by_col(' est_PctRural')
        se_rural = self.GS_NN.by_col(' se_PctRural')
        t_rural = self.GS_NN.by_col(' t_PctRural')
        est_pov = self.GS_NN.by_col(' est_PctPov')
        se_pov = self.GS_NN.by_col(' se_PctPov')
        t_pov = self.GS_NN.by_col(' t_PctPov')
        est_black = self.GS_NN.by_col(' est_PctBlack')
        se_black = self.GS_NN.by_col(' se_PctBlack')
        t_black = self.GS_NN.by_col(' t_PctBlack')
        yhat = self.GS_NN.by_col(' yhat')
        res = np.array(self.GS_NN.by_col(' residual'))
        std_res = np.array(self.GS_NN.by_col(' std_residual')).reshape((-1, 1))
        localR2 = np.array(self.GS_NN.by_col(' localR2')).reshape((-1, 1))
        inf = np.array(self.GS_NN.by_col(' influence')).reshape((-1, 1))
        cooksD = np.array(self.GS_NN.by_col(' CooksD')).reshape((-1, 1))

        model = GWR(self.coords, self.y, self.X, bw=49.000, kernel='gaussian',
                    fixed=False, sigma2_v1=False)

        rslt = model.fit()
        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)
        CV = get_CV(rslt)

        self.assertAlmostEquals(np.floor(AICc), 896.0)
        self.assertAlmostEquals(np.floor(AIC), 894.0)
        self.assertAlmostEquals(np.floor(BIC), 922.0)
        self.assertAlmostEquals(np.around(CV, 2), 17.91)
        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-04)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-04)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-04)
        np.testing.assert_allclose(est_rural, rslt.params[:, 1], rtol=1e-04)
        np.testing.assert_allclose(se_rural, rslt.bse[:, 1], rtol=1e-04)
        np.testing.assert_allclose(t_rural, rslt.tvalues[:, 1], rtol=1e-04)
        np.testing.assert_allclose(est_pov, rslt.params[:, 2], rtol=1e-04)
        np.testing.assert_allclose(se_pov, rslt.bse[:, 2], rtol=1e-04)
        np.testing.assert_allclose(t_pov, rslt.tvalues[:, 2], rtol=1e-04)
        np.testing.assert_allclose(est_black, rslt.params[:, 3], rtol=1e-02)
        np.testing.assert_allclose(se_black, rslt.bse[:, 3], rtol=1e-02)
        np.testing.assert_allclose(t_black, rslt.tvalues[:, 3], rtol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-05)
        np.testing.assert_allclose(res, rslt.resid_response, rtol=1e-04)
        np.testing.assert_allclose(std_res, rslt.std_res, rtol=1e-04)
        np.testing.assert_allclose(localR2, rslt.localR2, rtol=1e-05)
        np.testing.assert_allclose(inf, rslt.influ, rtol=1e-04)
        np.testing.assert_allclose(cooksD, rslt.cooksD, rtol=1e-00)

    def test_MGWR(self):
        std_y = (self.y - self.y.mean()) / self.y.std()
        std_X = (self.mgwr_X - self.mgwr_X.mean(axis=0)) / \
            self.mgwr_X.std(axis=0)
        selector = Sel_BW(self.coords, std_y, std_X, multi=True, constant=True)
        selector.search(multi_bw_min=[2], multi_bw_max=[159])
        model = MGWR(self.coords, std_y, std_X, selector=selector,
                     constant=True)

        rslt = model.fit()
        rslt_2 = model.fit(n_chunks=2)  #testing for n_chunks > 1
        rslt_3 = model.fit(n_chunks=3)
        rslt_20 = model.fit(n_chunks=20)

        model_hat = MGWR(self.coords, std_y, std_X, selector=selector,
                         constant=True, hat_matrix=True)
        rslt_hat = model_hat.fit()
        rslt_hat_2 = model_hat.fit(n_chunks=2)

        np.testing.assert_allclose(rslt_hat.R, rslt_hat_2.R, atol=1e-07)
        np.testing.assert_allclose(
            rslt_hat.S.dot(std_y).flatten(), self.MGWR.predy, atol=1e-07)

        varnames = ['X0', 'X1', 'X2', 'X3']

        # def suffixed(x):
        #     """ Quick anonymous function to suffix strings"""
        #     return ['_'.join(x) for x in varnames]

        np.testing.assert_allclose(rslt.predy.flatten(), self.MGWR.predy,
                                   atol=1e-07)
        np.testing.assert_allclose(rslt.params, self.MGWR[varnames].values,
                                   atol=1e-07)
        np.testing.assert_allclose(
            rslt.bse, self.MGWR[[s + "_bse" for s in varnames]].values,
            atol=1e-07)
        np.testing.assert_allclose(
            rslt_2.bse, self.MGWR[[s + "_bse" for s in varnames]].values,
            atol=1e-07)
        np.testing.assert_allclose(
            rslt_3.bse, self.MGWR[[s + "_bse" for s in varnames]].values,
            atol=1e-07)
        np.testing.assert_allclose(
            rslt_20.bse, self.MGWR[[s + "_bse" for s in varnames]].values,
            atol=1e-07)
        np.testing.assert_allclose(
            rslt.tvalues, self.MGWR[[s + "_tvalues" for s in varnames]].values,
            atol=1e-07)
        np.testing.assert_allclose(rslt.resid_response,
                                   self.MGWR.resid_response, atol=1e-04,
                                   rtol=1e-04)
        np.testing.assert_almost_equal(rslt.resid_ss, 50.899379467870425)
        np.testing.assert_almost_equal(rslt.aicc, 297.12013812258783)
        np.testing.assert_almost_equal(rslt.ENP, 11.36825087269831)
        np.testing.assert_allclose(rslt.ENP_j, [
            3.844671080264143, 3.513770805151652, 2.2580525278898254,
            1.7517564593926895
        ])
        np.testing.assert_allclose(rslt_2.ENP_j, [
            3.844671080264143, 3.513770805151652, 2.2580525278898254,
            1.7517564593926895
        ])
        np.testing.assert_allclose(rslt_3.ENP_j, [
            3.844671080264143, 3.513770805151652, 2.2580525278898254,
            1.7517564593926895
        ])
        np.testing.assert_allclose(rslt_20.ENP_j, [
            3.844671080264143, 3.513770805151652, 2.2580525278898254,
            1.7517564593926895
        ])
        np.testing.assert_allclose(
            rslt.adj_alpha_j,
            np.array([[0.02601003, 0.01300501, 0.0002601],
                      [0.02845945, 0.01422973, 0.00028459],
                      [0.04428595, 0.02214297, 0.00044286],
                      [0.05708556, 0.02854278, 0.00057086]]), atol=1e-07)
        np.testing.assert_allclose(
            rslt.critical_tval(),
            np.array([2.51210749, 2.47888792, 2.31069113, 2.21000184]),
            atol=1e-07)
        np.testing.assert_allclose(
            rslt.filter_tvals(),
            self.MGWR[[s + "_filter_tvalues" for s in varnames]].values,
            atol=1e-07)
        np.testing.assert_allclose(rslt.local_collinearity()[0].flatten(),
                                   self.MGWR.local_collinearity, atol=1e-07)

    def test_Prediction(self):
        coords = np.array(self.coords)
        index = np.arange(len(self.y))
        test = index[-10:]

        X_test = self.X[test]
        coords_test = coords[test]

        model = GWR(self.coords, self.y, self.X, 93, family=Gaussian(),
                    fixed=False, kernel='bisquare', sigma2_v1=False)
        results = model.predict(coords_test, X_test)

        params = np.array([
            22.77198, -0.10254, -0.215093, -0.01405, 19.10531, -0.094177,
            -0.232529, 0.071913, 19.743421, -0.080447, -0.30893, 0.083206,
            17.505759, -0.078919, -0.187955, 0.051719, 27.747402, -0.165335,
            -0.208553, 0.004067, 26.210627, -0.138398, -0.360514, 0.072199,
            18.034833, -0.077047, -0.260556, 0.084319, 28.452802, -0.163408,
            -0.14097, -0.063076, 22.353095, -0.103046, -0.226654, 0.002992,
            18.220508, -0.074034, -0.309812, 0.108636
        ]).reshape((10, 4))
        np.testing.assert_allclose(params, results.params, rtol=1e-03)

        bse = np.array([
            2.080166, 0.021462, 0.102954, 0.049627, 2.536355, 0.022111,
            0.123857, 0.051917, 1.967813, 0.019716, 0.102562, 0.054918,
            2.463219, 0.021745, 0.110297, 0.044189, 1.556056, 0.019513,
            0.12764, 0.040315, 1.664108, 0.020114, 0.131208, 0.041613, 2.5835,
            0.021481, 0.113158, 0.047243, 1.709483, 0.019752, 0.116944,
            0.043636, 1.958233, 0.020947, 0.09974, 0.049821, 2.276849,
            0.020122, 0.107867, 0.047842
        ]).reshape((10, 4))
        np.testing.assert_allclose(bse, results.bse, rtol=1e-03)

        tvalues = np.array([
            10.947193, -4.777659, -2.089223, -0.283103, 7.532584, -4.259179,
            -1.877395, 1.385161, 10.033179, -4.080362, -3.012133, 1.515096,
            7.106862, -3.629311, -1.704079, 1.17042, 17.831878, -8.473156,
            -1.633924, 0.100891, 15.750552, -6.880725, -2.74765, 1.734978,
            6.980774, -3.586757, -2.302575, 1.784818, 16.644095, -8.273001,
            -1.205451, -1.445501, 11.414933, -4.919384, -2.272458, 0.060064,
            8.00251, -3.679274, -2.872176, 2.270738
        ]).reshape((10, 4))
        np.testing.assert_allclose(tvalues, results.tvalues, rtol=1e-03)

        localR2 = np.array([[0.53068693], [0.59582647], [0.59700925],
                            [0.45769954], [0.54634509], [0.5494828],
                            [0.55159604], [0.55634237], [0.53903842],
                            [0.55884954]])
        np.testing.assert_allclose(localR2, results.localR2, rtol=1e-05)

        predictions = np.array([[10.51695514], [9.93321992], [8.92473026],
                                [5.47350219], [8.61756585], [12.8141851],
                                [5.55619405], [12.63004172], [8.70638418],
                                [8.17582599]])
        np.testing.assert_allclose(predictions, results.predictions,
                                   rtol=1e-05)

    def test_BS_NN_longlat(self):
        GA_longlat = os.path.join(
            os.path.dirname(__file__), 'ga_bs_nn_longlat_listwise.csv')
        self.BS_NN_longlat = io.open(GA_longlat)

        coords_longlat = list(
            zip(
                self.BS_NN_longlat.by_col(' x_coord'),
                self.BS_NN_longlat.by_col(' y_coord')))
        est_Int = self.BS_NN_longlat.by_col(' est_Intercept')
        se_Int = self.BS_NN_longlat.by_col(' se_Intercept')
        t_Int = self.BS_NN_longlat.by_col(' t_Intercept')
        est_rural = self.BS_NN_longlat.by_col(' est_PctRural')
        se_rural = self.BS_NN_longlat.by_col(' se_PctRural')
        t_rural = self.BS_NN_longlat.by_col(' t_PctRural')
        est_pov = self.BS_NN_longlat.by_col(' est_PctPov')
        se_pov = self.BS_NN_longlat.by_col(' se_PctPov')
        t_pov = self.BS_NN_longlat.by_col(' t_PctPov')
        est_black = self.BS_NN_longlat.by_col(' est_PctBlack')
        se_black = self.BS_NN_longlat.by_col(' se_PctBlack')
        t_black = self.BS_NN_longlat.by_col(' t_PctBlack')
        yhat = self.BS_NN_longlat.by_col(' yhat')
        res = np.array(self.BS_NN_longlat.by_col(' residual'))
        std_res = np.array(self.BS_NN_longlat.by_col(' std_residual')).reshape(
            (-1, 1))
        localR2 = np.array(self.BS_NN_longlat.by_col(' localR2')).reshape((-1,
                                                                           1))
        inf = np.array(self.BS_NN_longlat.by_col(' influence')).reshape((-1,
                                                                         1))
        cooksD = np.array(self.BS_NN_longlat.by_col(' CooksD')).reshape((-1,
                                                                         1))

        model = GWR(coords_longlat, self.y, self.X, bw=90.000, fixed=False,
                    spherical=True, sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)
        CV = get_CV(rslt)
        R2 = rslt.R2

        self.assertAlmostEquals(np.around(R2, 4), 0.5921)
        self.assertAlmostEquals(np.floor(AICc), 896.0)
        self.assertAlmostEquals(np.floor(AIC), 892.0)
        self.assertAlmostEquals(np.floor(BIC), 941.0)
        self.assertAlmostEquals(np.around(CV, 2), 19.11)
        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-04)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-04)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-04)
        np.testing.assert_allclose(est_rural, rslt.params[:, 1], rtol=1e-04)
        np.testing.assert_allclose(se_rural, rslt.bse[:, 1], rtol=1e-04)
        np.testing.assert_allclose(t_rural, rslt.tvalues[:, 1], rtol=1e-04)
        np.testing.assert_allclose(est_pov, rslt.params[:, 2], rtol=1e-04)
        np.testing.assert_allclose(se_pov, rslt.bse[:, 2], rtol=1e-04)
        np.testing.assert_allclose(t_pov, rslt.tvalues[:, 2], rtol=1e-04)
        np.testing.assert_allclose(est_black, rslt.params[:, 3], rtol=1e-02)
        np.testing.assert_allclose(se_black, rslt.bse[:, 3], rtol=1e-02)
        np.testing.assert_allclose(t_black, rslt.tvalues[:, 3], rtol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-05)
        np.testing.assert_allclose(res, rslt.resid_response, rtol=1e-04)
        np.testing.assert_allclose(std_res, rslt.std_res, rtol=1e-04)
        np.testing.assert_allclose(localR2, rslt.localR2, rtol=1e-05)
        np.testing.assert_allclose(inf, rslt.influ, rtol=1e-04)
        np.testing.assert_allclose(cooksD, rslt.cooksD, rtol=1e-00)


class TestGWRPoisson(unittest.TestCase):
    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__), 'tokyo/Tokyomortality.csv')
        data = io.open(data_path, mode='Ur')
        self.coords = list(
            zip(data.by_col('X_CENTROID'), data.by_col('Y_CENTROID')))
        self.y = np.array(data.by_col('db2564')).reshape((-1, 1))
        self.off = np.array(data.by_col('eb2564')).reshape((-1, 1))
        OCC = np.array(data.by_col('OCC_TEC')).reshape((-1, 1))
        OWN = np.array(data.by_col('OWNH')).reshape((-1, 1))
        POP = np.array(data.by_col('POP65')).reshape((-1, 1))
        UNEMP = np.array(data.by_col('UNEMP')).reshape((-1, 1))
        self.X = np.hstack([OCC, OWN, POP, UNEMP])
        self.BS_F = io.open(
            os.path.join(
                os.path.dirname(__file__), 'tokyo/tokyo_BS_F_listwise.csv'))
        self.BS_NN = io.open(
            os.path.join(
                os.path.dirname(__file__), 'tokyo/tokyo_BS_NN_listwise.csv'))
        self.GS_F = io.open(
            os.path.join(
                os.path.dirname(__file__), 'tokyo/tokyo_GS_F_listwise.csv'))
        self.GS_NN = io.open(
            os.path.join(
                os.path.dirname(__file__), 'tokyo/tokyo_GS_NN_listwise.csv'))
        self.BS_NN_OFF = io.open(
            os.path.join(
                os.path.dirname(__file__),
                'tokyo/tokyo_BS_NN_OFF_listwise.csv'))

    def test_BS_F(self):
        est_Int = self.BS_F.by_col(' est_Intercept')
        se_Int = self.BS_F.by_col(' se_Intercept')
        t_Int = self.BS_F.by_col(' t_Intercept')
        est_OCC = self.BS_F.by_col(' est_OCC_TEC')
        se_OCC = self.BS_F.by_col(' se_OCC_TEC')
        t_OCC = self.BS_F.by_col(' t_OCC_TEC')
        est_OWN = self.BS_F.by_col(' est_OWNH')
        se_OWN = self.BS_F.by_col(' se_OWNH')
        t_OWN = self.BS_F.by_col(' t_OWNH')
        est_POP = self.BS_F.by_col(' est_POP65')
        se_POP = self.BS_F.by_col(' se_POP65')
        t_POP = self.BS_F.by_col(' t_POP65')
        est_UNEMP = self.BS_F.by_col(' est_UNEMP')
        se_UNEMP = self.BS_F.by_col(' se_UNEMP')
        t_UNEMP = self.BS_F.by_col(' t_UNEMP')
        yhat = self.BS_F.by_col(' yhat')
        pdev = np.array(self.BS_F.by_col(' localpdev')).reshape((-1, 1))

        model = GWR(self.coords, self.y, self.X, bw=26029.625,
                    family=Poisson(), kernel='bisquare', fixed=True,
                    sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)

        self.assertAlmostEquals(np.floor(AICc), 13294.0)
        self.assertAlmostEquals(np.floor(AIC), 13247.0)
        self.assertAlmostEquals(np.floor(BIC), 13485.0)

        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-05)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-03)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-03)
        np.testing.assert_allclose(est_OCC, rslt.params[:, 1], rtol=1e-04)
        np.testing.assert_allclose(se_OCC, rslt.bse[:, 1], rtol=1e-02)
        np.testing.assert_allclose(t_OCC, rslt.tvalues[:, 1], rtol=1e-02)
        np.testing.assert_allclose(est_OWN, rslt.params[:, 2], rtol=1e-04)
        np.testing.assert_allclose(se_OWN, rslt.bse[:, 2], rtol=1e-03)
        np.testing.assert_allclose(t_OWN, rslt.tvalues[:, 2], rtol=1e-03)
        np.testing.assert_allclose(est_POP, rslt.params[:, 3], rtol=1e-04)
        np.testing.assert_allclose(se_POP, rslt.bse[:, 3], rtol=1e-02)
        np.testing.assert_allclose(t_POP, rslt.tvalues[:, 3], rtol=1e-02)
        np.testing.assert_allclose(est_UNEMP, rslt.params[:, 4], rtol=1e-04)
        np.testing.assert_allclose(se_UNEMP, rslt.bse[:, 4], rtol=1e-02)
        np.testing.assert_allclose(t_UNEMP, rslt.tvalues[:, 4], rtol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-05)
        np.testing.assert_allclose(pdev, rslt.pDev, rtol=1e-05)

    def test_BS_NN(self):
        est_Int = self.BS_NN.by_col(' est_Intercept')
        se_Int = self.BS_NN.by_col(' se_Intercept')
        t_Int = self.BS_NN.by_col(' t_Intercept')
        est_OCC = self.BS_NN.by_col(' est_OCC_TEC')
        se_OCC = self.BS_NN.by_col(' se_OCC_TEC')
        t_OCC = self.BS_NN.by_col(' t_OCC_TEC')
        est_OWN = self.BS_NN.by_col(' est_OWNH')
        se_OWN = self.BS_NN.by_col(' se_OWNH')
        t_OWN = self.BS_NN.by_col(' t_OWNH')
        est_POP = self.BS_NN.by_col(' est_POP65')
        se_POP = self.BS_NN.by_col(' se_POP65')
        t_POP = self.BS_NN.by_col(' t_POP65')
        est_UNEMP = self.BS_NN.by_col(' est_UNEMP')
        se_UNEMP = self.BS_NN.by_col(' se_UNEMP')
        t_UNEMP = self.BS_NN.by_col(' t_UNEMP')
        yhat = self.BS_NN.by_col(' yhat')
        pdev = np.array(self.BS_NN.by_col(' localpdev')).reshape((-1, 1))

        model = GWR(self.coords, self.y, self.X, bw=50, family=Poisson(),
                    kernel='bisquare', fixed=False, sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)
        D2 = rslt.D2

        self.assertAlmostEquals(np.floor(AICc), 13285)
        self.assertAlmostEquals(np.floor(AIC), 13259.0)
        self.assertAlmostEquals(np.floor(BIC), 13442.0)
        self.assertAlmostEquals(np.round(D2, 3), 0.747)

        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-04)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-02)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-02)
        np.testing.assert_allclose(est_OCC, rslt.params[:, 1], rtol=1e-03)
        np.testing.assert_allclose(se_OCC, rslt.bse[:, 1], rtol=1e-02)
        np.testing.assert_allclose(t_OCC, rslt.tvalues[:, 1], rtol=1e-02)
        np.testing.assert_allclose(est_OWN, rslt.params[:, 2], rtol=1e-04)
        np.testing.assert_allclose(se_OWN, rslt.bse[:, 2], rtol=1e-02)
        np.testing.assert_allclose(t_OWN, rslt.tvalues[:, 2], rtol=1e-02)
        np.testing.assert_allclose(est_POP, rslt.params[:, 3], rtol=1e-03)
        np.testing.assert_allclose(se_POP, rslt.bse[:, 3], rtol=1e-02)
        np.testing.assert_allclose(t_POP, rslt.tvalues[:, 3], rtol=1e-02)
        np.testing.assert_allclose(est_UNEMP, rslt.params[:, 4], rtol=1e-04)
        np.testing.assert_allclose(se_UNEMP, rslt.bse[:, 4], rtol=1e-02)
        np.testing.assert_allclose(t_UNEMP, rslt.tvalues[:, 4], rtol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-04)
        np.testing.assert_allclose(pdev, rslt.pDev, rtol=1e-05)

    def test_BS_NN_Offset(self):
        est_Int = self.BS_NN_OFF.by_col(' est_Intercept')
        se_Int = self.BS_NN_OFF.by_col(' se_Intercept')
        t_Int = self.BS_NN_OFF.by_col(' t_Intercept')
        est_OCC = self.BS_NN_OFF.by_col(' est_OCC_TEC')
        se_OCC = self.BS_NN_OFF.by_col(' se_OCC_TEC')
        t_OCC = self.BS_NN_OFF.by_col(' t_OCC_TEC')
        est_OWN = self.BS_NN_OFF.by_col(' est_OWNH')
        se_OWN = self.BS_NN_OFF.by_col(' se_OWNH')
        t_OWN = self.BS_NN_OFF.by_col(' t_OWNH')
        est_POP = self.BS_NN_OFF.by_col(' est_POP65')
        se_POP = self.BS_NN_OFF.by_col(' se_POP65')
        t_POP = self.BS_NN_OFF.by_col(' t_POP65')
        est_UNEMP = self.BS_NN_OFF.by_col(' est_UNEMP')
        se_UNEMP = self.BS_NN_OFF.by_col(' se_UNEMP')
        t_UNEMP = self.BS_NN_OFF.by_col(' t_UNEMP')
        yhat = self.BS_NN_OFF.by_col(' yhat')
        pdev = np.array(self.BS_NN_OFF.by_col(' localpdev')).reshape((-1, 1))

        model = GWR(self.coords, self.y, self.X, bw=100, offset=self.off,
                    family=Poisson(), kernel='bisquare', fixed=False,
                    sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)
        D2 = rslt.D2

        self.assertAlmostEquals(np.floor(AICc), 367.0)
        self.assertAlmostEquals(np.floor(AIC), 361.0)
        self.assertAlmostEquals(np.floor(BIC), 451.0)
        self.assertAlmostEquals(np.round(D2, 3), 0.676)

        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-02,
                                   atol=1e-02)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-02,
                                   atol=1e-02)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-01,
                                   atol=1e-02)
        np.testing.assert_allclose(est_OCC, rslt.params[:, 1], rtol=1e-03,
                                   atol=1e-02)
        np.testing.assert_allclose(se_OCC, rslt.bse[:, 1], rtol=1e-02,
                                   atol=1e-02)
        np.testing.assert_allclose(t_OCC, rslt.tvalues[:, 1], rtol=1e-01,
                                   atol=1e-02)
        np.testing.assert_allclose(est_OWN, rslt.params[:, 2], rtol=1e-04,
                                   atol=1e-02)
        np.testing.assert_allclose(se_OWN, rslt.bse[:, 2], rtol=1e-02,
                                   atol=1e-02)
        np.testing.assert_allclose(t_OWN, rslt.tvalues[:, 2], rtol=1e-01,
                                   atol=1e-02)
        np.testing.assert_allclose(est_POP, rslt.params[:, 3], rtol=1e-03,
                                   atol=1e-02)
        np.testing.assert_allclose(se_POP, rslt.bse[:, 3], rtol=1e-02,
                                   atol=1e-02)
        np.testing.assert_allclose(t_POP, rslt.tvalues[:, 3], rtol=1e-01,
                                   atol=1e-02)
        np.testing.assert_allclose(est_UNEMP, rslt.params[:, 4], rtol=1e-04,
                                   atol=1e-02)
        np.testing.assert_allclose(se_UNEMP, rslt.bse[:, 4], rtol=1e-02,
                                   atol=1e-02)
        np.testing.assert_allclose(t_UNEMP, rslt.tvalues[:, 4], rtol=1e-01,
                                   atol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-03, atol=1e-02)
        np.testing.assert_allclose(pdev, rslt.pDev, rtol=1e-04, atol=1e-02)

    def test_GS_F(self):
        est_Int = self.GS_F.by_col(' est_Intercept')
        se_Int = self.GS_F.by_col(' se_Intercept')
        t_Int = self.GS_F.by_col(' t_Intercept')
        est_OCC = self.GS_F.by_col(' est_OCC_TEC')
        se_OCC = self.GS_F.by_col(' se_OCC_TEC')
        t_OCC = self.GS_F.by_col(' t_OCC_TEC')
        est_OWN = self.GS_F.by_col(' est_OWNH')
        se_OWN = self.GS_F.by_col(' se_OWNH')
        t_OWN = self.GS_F.by_col(' t_OWNH')
        est_POP = self.GS_F.by_col(' est_POP65')
        se_POP = self.GS_F.by_col(' se_POP65')
        t_POP = self.GS_F.by_col(' t_POP65')
        est_UNEMP = self.GS_F.by_col(' est_UNEMP')
        se_UNEMP = self.GS_F.by_col(' se_UNEMP')
        t_UNEMP = self.GS_F.by_col(' t_UNEMP')
        yhat = self.GS_F.by_col(' yhat')
        pdev = np.array(self.GS_F.by_col(' localpdev')).reshape((-1, 1))

        model = GWR(self.coords, self.y, self.X, bw=8764.474, family=Poisson(),
                    kernel='gaussian', fixed=True, sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)

        self.assertAlmostEquals(np.floor(AICc), 11283.0)
        self.assertAlmostEquals(np.floor(AIC), 11211.0)
        self.assertAlmostEquals(np.floor(BIC), 11497.0)
        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-03)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-02)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-02)
        np.testing.assert_allclose(est_OCC, rslt.params[:, 1], rtol=1e-03)
        np.testing.assert_allclose(se_OCC, rslt.bse[:, 1], rtol=1e-02)
        np.testing.assert_allclose(t_OCC, rslt.tvalues[:, 1], rtol=1e-02)
        np.testing.assert_allclose(est_OWN, rslt.params[:, 2], rtol=1e-03)
        np.testing.assert_allclose(se_OWN, rslt.bse[:, 2], rtol=1e-02)
        np.testing.assert_allclose(t_OWN, rslt.tvalues[:, 2], rtol=1e-02)
        np.testing.assert_allclose(est_POP, rslt.params[:, 3], rtol=1e-02)
        np.testing.assert_allclose(se_POP, rslt.bse[:, 3], rtol=1e-02)
        np.testing.assert_allclose(t_POP, rslt.tvalues[:, 3], rtol=1e-02)
        np.testing.assert_allclose(est_UNEMP, rslt.params[:, 4], rtol=1e-02)
        np.testing.assert_allclose(se_UNEMP, rslt.bse[:, 4], rtol=1e-02)
        np.testing.assert_allclose(t_UNEMP, rslt.tvalues[:, 4], rtol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-04)
        np.testing.assert_allclose(pdev, rslt.pDev, rtol=1e-05)

    def test_GS_NN(self):
        est_Int = self.GS_NN.by_col(' est_Intercept')
        se_Int = self.GS_NN.by_col(' se_Intercept')
        t_Int = self.GS_NN.by_col(' t_Intercept')
        est_OCC = self.GS_NN.by_col(' est_OCC_TEC')
        se_OCC = self.GS_NN.by_col(' se_OCC_TEC')
        t_OCC = self.GS_NN.by_col(' t_OCC_TEC')
        est_OWN = self.GS_NN.by_col(' est_OWNH')
        se_OWN = self.GS_NN.by_col(' se_OWNH')
        t_OWN = self.GS_NN.by_col(' t_OWNH')
        est_POP = self.GS_NN.by_col(' est_POP65')
        se_POP = self.GS_NN.by_col(' se_POP65')
        t_POP = self.GS_NN.by_col(' t_POP65')
        est_UNEMP = self.GS_NN.by_col(' est_UNEMP')
        se_UNEMP = self.GS_NN.by_col(' se_UNEMP')
        t_UNEMP = self.GS_NN.by_col(' t_UNEMP')
        yhat = self.GS_NN.by_col(' yhat')
        pdev = np.array(self.GS_NN.by_col(' localpdev')).reshape((-1, 1))

        model = GWR(self.coords, self.y, self.X, bw=50, family=Poisson(),
                    kernel='gaussian', fixed=False, sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)

        self.assertAlmostEquals(np.floor(AICc), 21070.0)
        self.assertAlmostEquals(np.floor(AIC), 21069.0)
        self.assertAlmostEquals(np.floor(BIC), 21111.0)
        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-04)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-02)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-02)
        np.testing.assert_allclose(est_OCC, rslt.params[:, 1], rtol=1e-03)
        np.testing.assert_allclose(se_OCC, rslt.bse[:, 1], rtol=1e-02)
        np.testing.assert_allclose(t_OCC, rslt.tvalues[:, 1], rtol=1e-02)
        np.testing.assert_allclose(est_OWN, rslt.params[:, 2], rtol=1e-04)
        np.testing.assert_allclose(se_OWN, rslt.bse[:, 2], rtol=1e-02)
        np.testing.assert_allclose(t_OWN, rslt.tvalues[:, 2], rtol=1e-02)
        np.testing.assert_allclose(est_POP, rslt.params[:, 3], rtol=1e-02)
        np.testing.assert_allclose(se_POP, rslt.bse[:, 3], rtol=1e-02)
        np.testing.assert_allclose(t_POP, rslt.tvalues[:, 3], rtol=1e-02)
        np.testing.assert_allclose(est_UNEMP, rslt.params[:, 4], rtol=1e-02)
        np.testing.assert_allclose(se_UNEMP, rslt.bse[:, 4], rtol=1e-02)
        np.testing.assert_allclose(t_UNEMP, rslt.tvalues[:, 4], rtol=1e-02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-04)
        np.testing.assert_allclose(pdev, rslt.pDev, rtol=1e-05)


class TestGWRBinomial(unittest.TestCase):
    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__), 'clearwater/landslides.csv')
        data = io.open(data_path)
        self.coords = list(zip(data.by_col('X'), data.by_col('Y')))
        self.y = np.array(data.by_col('Landslid')).reshape((-1, 1))
        ELEV = np.array(data.by_col('Elev')).reshape((-1, 1))
        SLOPE = np.array(data.by_col('Slope')).reshape((-1, 1))
        SIN = np.array(data.by_col('SinAspct')).reshape((-1, 1))
        COS = np.array(data.by_col('CosAspct')).reshape((-1, 1))
        SOUTH = np.array(data.by_col('AbsSouth')).reshape((-1, 1))
        DIST = np.array(data.by_col('DistStrm')).reshape((-1, 1))
        self.X = np.hstack([ELEV, SLOPE, SIN, COS, SOUTH, DIST])
        self.BS_F = io.open(
            os.path.join(
                os.path.dirname(__file__),
                'clearwater/clearwater_BS_F_listwise.csv'))
        self.BS_NN = io.open(
            os.path.join(
                os.path.dirname(__file__),
                'clearwater/clearwater_BS_NN_listwise.csv'))
        self.GS_F = io.open(
            os.path.join(
                os.path.dirname(__file__),
                'clearwater/clearwater_GS_F_listwise.csv'))
        self.GS_NN = io.open(
            os.path.join(
                os.path.dirname(__file__),
                'clearwater/clearwater_GS_NN_listwise.csv'))

    def test_BS_F(self):
        est_Int = self.BS_F.by_col(' est_Intercept')
        se_Int = self.BS_F.by_col(' se_Intercept')
        t_Int = self.BS_F.by_col(' t_Intercept')
        est_elev = self.BS_F.by_col(' est_Elev')
        se_elev = self.BS_F.by_col(' se_Elev')
        t_elev = self.BS_F.by_col(' t_Elev')
        est_slope = self.BS_F.by_col(' est_Slope')
        se_slope = self.BS_F.by_col(' se_Slope')
        t_slope = self.BS_F.by_col(' t_Slope')
        est_sin = self.BS_F.by_col(' est_SinAspct')
        se_sin = self.BS_F.by_col(' se_SinAspct')
        t_sin = self.BS_F.by_col(' t_SinAspct')
        est_cos = self.BS_F.by_col(' est_CosAspct')
        se_cos = self.BS_F.by_col(' se_CosAspct')
        t_cos = self.BS_F.by_col(' t_CosAspct')
        est_south = self.BS_F.by_col(' est_AbsSouth')
        se_south = self.BS_F.by_col(' se_AbsSouth')
        t_south = self.BS_F.by_col(' t_AbsSouth')
        est_strm = self.BS_F.by_col(' est_DistStrm')
        se_strm = self.BS_F.by_col(' se_DistStrm')
        t_strm = self.BS_F.by_col(' t_DistStrm')
        yhat = self.BS_F.by_col(' yhat')
        pdev = np.array(self.BS_F.by_col(' localpdev')).reshape((-1, 1))

        model = GWR(self.coords, self.y, self.X, bw=19642.170,
                    family=Binomial(), kernel='bisquare', fixed=True,
                    sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)

        self.assertAlmostEquals(np.floor(AICc), 275.0)
        self.assertAlmostEquals(np.floor(AIC), 271.0)
        self.assertAlmostEquals(np.floor(BIC), 349.0)
        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-00)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-00)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-00)
        np.testing.assert_allclose(est_elev, rslt.params[:, 1], rtol=1e-00)
        np.testing.assert_allclose(se_elev, rslt.bse[:, 1], rtol=1e-00)
        np.testing.assert_allclose(t_elev, rslt.tvalues[:, 1], rtol=1e-00)
        np.testing.assert_allclose(est_slope, rslt.params[:, 2], rtol=1e-00)
        np.testing.assert_allclose(se_slope, rslt.bse[:, 2], rtol=1e-00)
        np.testing.assert_allclose(t_slope, rslt.tvalues[:, 2], rtol=1e-00)
        np.testing.assert_allclose(est_sin, rslt.params[:, 3], rtol=1e01)
        np.testing.assert_allclose(se_sin, rslt.bse[:, 3], rtol=1e01)
        np.testing.assert_allclose(t_sin, rslt.tvalues[:, 3], rtol=1e01)
        np.testing.assert_allclose(est_cos, rslt.params[:, 4], rtol=1e01)
        np.testing.assert_allclose(se_cos, rslt.bse[:, 4], rtol=1e01)
        np.testing.assert_allclose(t_cos, rslt.tvalues[:, 4], rtol=1e01)
        np.testing.assert_allclose(est_south, rslt.params[:, 5], rtol=1e01)
        np.testing.assert_allclose(se_south, rslt.bse[:, 5], rtol=1e01)
        np.testing.assert_allclose(t_south, rslt.tvalues[:, 5], rtol=1e01)
        np.testing.assert_allclose(est_strm, rslt.params[:, 6], rtol=1e02)
        np.testing.assert_allclose(se_strm, rslt.bse[:, 6], rtol=1e01)
        np.testing.assert_allclose(t_strm, rslt.tvalues[:, 6], rtol=1e02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-01)
        # This test fails - likely due to compound rounding errors
        # Has been tested using statsmodels.family calculations and
        # code from Jing's python version, which both yield the same
        #np.testing.assert_allclose(pdev, rslt.pDev, rtol=1e-05)

    def test_BS_NN(self):
        est_Int = self.BS_NN.by_col(' est_Intercept')
        se_Int = self.BS_NN.by_col(' se_Intercept')
        t_Int = self.BS_NN.by_col(' t_Intercept')
        est_elev = self.BS_NN.by_col(' est_Elev')
        se_elev = self.BS_NN.by_col(' se_Elev')
        t_elev = self.BS_NN.by_col(' t_Elev')
        est_slope = self.BS_NN.by_col(' est_Slope')
        se_slope = self.BS_NN.by_col(' se_Slope')
        t_slope = self.BS_NN.by_col(' t_Slope')
        est_sin = self.BS_NN.by_col(' est_SinAspct')
        se_sin = self.BS_NN.by_col(' se_SinAspct')
        t_sin = self.BS_NN.by_col(' t_SinAspct')
        est_cos = self.BS_NN.by_col(' est_CosAspct')
        se_cos = self.BS_NN.by_col(' se_CosAspct')
        t_cos = self.BS_NN.by_col(' t_CosAspct')
        est_south = self.BS_NN.by_col(' est_AbsSouth')
        se_south = self.BS_NN.by_col(' se_AbsSouth')
        t_south = self.BS_NN.by_col(' t_AbsSouth')
        est_strm = self.BS_NN.by_col(' est_DistStrm')
        se_strm = self.BS_NN.by_col(' se_DistStrm')
        t_strm = self.BS_NN.by_col(' t_DistStrm')
        yhat = self.BS_NN.by_col(' yhat')
        pdev = self.BS_NN.by_col(' localpdev')

        model = GWR(self.coords, self.y, self.X, bw=158, family=Binomial(),
                    kernel='bisquare', fixed=False, sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)
        D2 = rslt.D2

        self.assertAlmostEquals(np.floor(AICc), 277.0)
        self.assertAlmostEquals(np.floor(AIC), 271.0)
        self.assertAlmostEquals(np.floor(BIC), 358.0)
        self.assertAlmostEquals(np.round(D2, 3), 0.319)

        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-00)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-00)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-00)
        np.testing.assert_allclose(est_elev, rslt.params[:, 1], rtol=1e-00)
        np.testing.assert_allclose(se_elev, rslt.bse[:, 1], rtol=1e-00)
        np.testing.assert_allclose(t_elev, rslt.tvalues[:, 1], rtol=1e-00)
        np.testing.assert_allclose(est_slope, rslt.params[:, 2], rtol=1e-00)
        np.testing.assert_allclose(se_slope, rslt.bse[:, 2], rtol=1e-00)
        np.testing.assert_allclose(t_slope, rslt.tvalues[:, 2], rtol=1e-00)
        np.testing.assert_allclose(est_sin, rslt.params[:, 3], rtol=1e01)
        np.testing.assert_allclose(se_sin, rslt.bse[:, 3], rtol=1e01)
        np.testing.assert_allclose(t_sin, rslt.tvalues[:, 3], rtol=1e01)
        np.testing.assert_allclose(est_cos, rslt.params[:, 4], rtol=1e01)
        np.testing.assert_allclose(se_cos, rslt.bse[:, 4], rtol=1e01)
        np.testing.assert_allclose(t_cos, rslt.tvalues[:, 4], rtol=1e01)
        np.testing.assert_allclose(est_south, rslt.params[:, 5], rtol=1e01)
        np.testing.assert_allclose(se_south, rslt.bse[:, 5], rtol=1e01)
        np.testing.assert_allclose(t_south, rslt.tvalues[:, 5], rtol=1e01)
        np.testing.assert_allclose(est_strm, rslt.params[:, 6], rtol=1e03)
        np.testing.assert_allclose(se_strm, rslt.bse[:, 6], rtol=1e01)
        np.testing.assert_allclose(t_strm, rslt.tvalues[:, 6], rtol=1e03)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-01)
        # This test fails - likely due to compound rounding errors
        # Has been tested using statsmodels.family calculations and
        # code from Jing's python version, which both yield the same
        #np.testing.assert_allclose(pdev, rslt.pDev, rtol=1e-05)

    def test_GS_F(self):
        est_Int = self.GS_F.by_col(' est_Intercept')
        se_Int = self.GS_F.by_col(' se_Intercept')
        t_Int = self.GS_F.by_col(' t_Intercept')
        est_elev = self.GS_F.by_col(' est_Elev')
        se_elev = self.GS_F.by_col(' se_Elev')
        t_elev = self.GS_F.by_col(' t_Elev')
        est_slope = self.GS_F.by_col(' est_Slope')
        se_slope = self.GS_F.by_col(' se_Slope')
        t_slope = self.GS_F.by_col(' t_Slope')
        est_sin = self.GS_F.by_col(' est_SinAspct')
        se_sin = self.GS_F.by_col(' se_SinAspct')
        t_sin = self.GS_F.by_col(' t_SinAspct')
        est_cos = self.GS_F.by_col(' est_CosAspct')
        se_cos = self.GS_F.by_col(' se_CosAspct')
        t_cos = self.GS_F.by_col(' t_CosAspct')
        est_south = self.GS_F.by_col(' est_AbsSouth')
        se_south = self.GS_F.by_col(' se_AbsSouth')
        t_south = self.GS_F.by_col(' t_AbsSouth')
        est_strm = self.GS_F.by_col(' est_DistStrm')
        se_strm = self.GS_F.by_col(' se_DistStrm')
        t_strm = self.GS_F.by_col(' t_DistStrm')
        yhat = self.GS_F.by_col(' yhat')
        pdev = self.GS_F.by_col(' localpdev')

        model = GWR(self.coords, self.y, self.X, bw=8929.061,
                    family=Binomial(), kernel='gaussian', fixed=True,
                    sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)

        self.assertAlmostEquals(np.floor(AICc), 276.0)
        self.assertAlmostEquals(np.floor(AIC), 272.0)
        self.assertAlmostEquals(np.floor(BIC), 341.0)
        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-00)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-00)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-00)
        np.testing.assert_allclose(est_elev, rslt.params[:, 1], rtol=1e-00)
        np.testing.assert_allclose(se_elev, rslt.bse[:, 1], rtol=1e-00)
        np.testing.assert_allclose(t_elev, rslt.tvalues[:, 1], rtol=1e-00)
        np.testing.assert_allclose(est_slope, rslt.params[:, 2], rtol=1e-00)
        np.testing.assert_allclose(se_slope, rslt.bse[:, 2], rtol=1e-00)
        np.testing.assert_allclose(t_slope, rslt.tvalues[:, 2], rtol=1e-00)
        np.testing.assert_allclose(est_sin, rslt.params[:, 3], rtol=1e01)
        np.testing.assert_allclose(se_sin, rslt.bse[:, 3], rtol=1e01)
        np.testing.assert_allclose(t_sin, rslt.tvalues[:, 3], rtol=1e01)
        np.testing.assert_allclose(est_cos, rslt.params[:, 4], rtol=1e01)
        np.testing.assert_allclose(se_cos, rslt.bse[:, 4], rtol=1e01)
        np.testing.assert_allclose(t_cos, rslt.tvalues[:, 4], rtol=1e01)
        np.testing.assert_allclose(est_south, rslt.params[:, 5], rtol=1e01)
        np.testing.assert_allclose(se_south, rslt.bse[:, 5], rtol=1e01)
        np.testing.assert_allclose(t_south, rslt.tvalues[:, 5], rtol=1e01)
        np.testing.assert_allclose(est_strm, rslt.params[:, 6], rtol=1e02)
        np.testing.assert_allclose(se_strm, rslt.bse[:, 6], rtol=1e01)
        np.testing.assert_allclose(t_strm, rslt.tvalues[:, 6], rtol=1e02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-01)
        # This test fails - likely due to compound rounding errors
        # Has been tested using statsmodels.family calculations and
        # code from Jing's python version, which both yield the same
        #np.testing.assert_allclose(pdev, rslt.pDev, rtol=1e-05)

    def test_GS_NN(self):
        est_Int = self.GS_NN.by_col(' est_Intercept')
        se_Int = self.GS_NN.by_col(' se_Intercept')
        t_Int = self.GS_NN.by_col(' t_Intercept')
        est_elev = self.GS_NN.by_col(' est_Elev')
        se_elev = self.GS_NN.by_col(' se_Elev')
        t_elev = self.GS_NN.by_col(' t_Elev')
        est_slope = self.GS_NN.by_col(' est_Slope')
        se_slope = self.GS_NN.by_col(' se_Slope')
        t_slope = self.GS_NN.by_col(' t_Slope')
        est_sin = self.GS_NN.by_col(' est_SinAspct')
        se_sin = self.GS_NN.by_col(' se_SinAspct')
        t_sin = self.GS_NN.by_col(' t_SinAspct')
        est_cos = self.GS_NN.by_col(' est_CosAspct')
        se_cos = self.GS_NN.by_col(' se_CosAspct')
        t_cos = self.GS_NN.by_col(' t_CosAspct')
        est_south = self.GS_NN.by_col(' est_AbsSouth')
        se_south = self.GS_NN.by_col(' se_AbsSouth')
        t_south = self.GS_NN.by_col(' t_AbsSouth')
        est_strm = self.GS_NN.by_col(' est_DistStrm')
        se_strm = self.GS_NN.by_col(' se_DistStrm')
        t_strm = self.GS_NN.by_col(' t_DistStrm')
        yhat = self.GS_NN.by_col(' yhat')
        pdev = self.GS_NN.by_col(' localpdev')

        model = GWR(self.coords, self.y, self.X, bw=64, family=Binomial(),
                    kernel='gaussian', fixed=False, sigma2_v1=False)
        rslt = model.fit()

        AICc = get_AICc(rslt)
        AIC = get_AIC(rslt)
        BIC = get_BIC(rslt)

        self.assertAlmostEquals(np.floor(AICc), 276.0)
        self.assertAlmostEquals(np.floor(AIC), 273.0)
        self.assertAlmostEquals(np.floor(BIC), 331.0)
        np.testing.assert_allclose(est_Int, rslt.params[:, 0], rtol=1e-00)
        np.testing.assert_allclose(se_Int, rslt.bse[:, 0], rtol=1e-00)
        np.testing.assert_allclose(t_Int, rslt.tvalues[:, 0], rtol=1e-00)
        np.testing.assert_allclose(est_elev, rslt.params[:, 1], rtol=1e-00)
        np.testing.assert_allclose(se_elev, rslt.bse[:, 1], rtol=1e-00)
        np.testing.assert_allclose(t_elev, rslt.tvalues[:, 1], rtol=1e-00)
        np.testing.assert_allclose(est_slope, rslt.params[:, 2], rtol=1e-00)
        np.testing.assert_allclose(se_slope, rslt.bse[:, 2], rtol=1e-00)
        np.testing.assert_allclose(t_slope, rslt.tvalues[:, 2], rtol=1e-00)
        np.testing.assert_allclose(est_sin, rslt.params[:, 3], rtol=1e01)
        np.testing.assert_allclose(se_sin, rslt.bse[:, 3], rtol=1e01)
        np.testing.assert_allclose(t_sin, rslt.tvalues[:, 3], rtol=1e01)
        np.testing.assert_allclose(est_cos, rslt.params[:, 4], rtol=1e01)
        np.testing.assert_allclose(se_cos, rslt.bse[:, 4], rtol=1e01)
        np.testing.assert_allclose(t_cos, rslt.tvalues[:, 4], rtol=1e01)
        np.testing.assert_allclose(est_south, rslt.params[:, 5], rtol=1e01)
        np.testing.assert_allclose(se_south, rslt.bse[:, 5], rtol=1e01)
        np.testing.assert_allclose(t_south, rslt.tvalues[:, 5], rtol=1e01)
        np.testing.assert_allclose(est_strm, rslt.params[:, 6], rtol=1e02)
        np.testing.assert_allclose(se_strm, rslt.bse[:, 6], rtol=1e01)
        np.testing.assert_allclose(t_strm, rslt.tvalues[:, 6], rtol=1e02)
        np.testing.assert_allclose(yhat, rslt.mu, rtol=1e-00)
        # This test fails - likely due to compound rounding errors
        # Has been tested using statsmodels.family calculations and
        # code from Jing's python version, which both yield the same
        #np.testing.assert_allclose(pdev, rslt.pDev, rtol=1e-05)


if __name__ == '__main__':
    unittest.main()



"""
Diagnostics for estimated gwr models
"""
__author__ = "Taylor Oshan tayoshan@gmail.com"

import numpy as np
from scipy import linalg
from spglm.family import Gaussian, Poisson, Binomial


def get_AICc(gwr):
    """
    Get AICc value
    
    Gaussian: p61, (2.33), Fotheringham, Brunsdon and Charlton (2002)
    
    GWGLM: AICc=AIC+2k(k+1)/(n-k-1), Nakaya et al. (2005): p2704, (36)

    """
    n = gwr.n
    k = gwr.tr_S
    #sigma2 = gwr.sigma2
    if isinstance(gwr.family, Gaussian):
        aicc = -2.0 * gwr.llf + 2.0 * n * (k + 1.0) / (
            n - k - 2.0)  #equivalent to below but
        #can't control denominator of sigma without altering GLM familt code
        #aicc = n*np.log(sigma2) + n*np.log(2.0*np.pi) + n*(n+k)/(n-k-2.0)
    elif isinstance(gwr.family, (Poisson, Binomial)):
        aicc = get_AIC(gwr) + 2.0 * k * (k + 1.0) / (n - k - 1.0)
    return aicc


def get_AIC(gwr):
    """
    Get AIC calue

    Gaussian: p96, (4.22), Fotheringham, Brunsdon and Charlton (2002)

    GWGLM:  AIC(G)=D(G) + 2K(G), where D and K denote the deviance and the effective
    number of parameters in the model with bandwidth G, respectively.
    
    """
    k = gwr.tr_S
    #deviance = -2*log-likelihood
    y = gwr.y
    mu = gwr.mu
    if isinstance(gwr.family, Gaussian):
        aic = -2.0 * gwr.llf + 2.0 * (k + 1)
    elif isinstance(gwr.family, (Poisson, Binomial)):
        aic = np.sum(gwr.family.resid_dev(y, mu)**2) + 2.0 * k
    return aic


def get_BIC(gwr):
    """
    Get BIC value

    Gaussian: p61 (2.34), Fotheringham, Brunsdon and Charlton (2002)
    BIC = -2log(L)+klog(n)

    GWGLM: BIC = dev + tr_S * log(n)

    """
    n = gwr.n  # (scalar) number of observations
    k = gwr.tr_S
    y = gwr.y
    mu = gwr.mu
    if isinstance(gwr.family, Gaussian):
        bic = -2.0 * gwr.llf + (k + 1) * np.log(n)
    elif isinstance(gwr.family, (Poisson, Binomial)):
        bic = np.sum(gwr.family.resid_dev(y, mu)**2) + k * np.log(n)
    return bic


def get_CV(gwr):
    """
    Get CV value

    Gaussian only

    Methods: p60, (2.31) or p212 (9.4)
    Fotheringham, A. S., Brunsdon, C., & Charlton, M. (2002).
    Geographically weighted regression: the analysis of spatially varying relationships.
    Modification: sum of residual squared is divided by n according to GWR4 results

    """
    aa = gwr.resid_response.reshape((-1, 1)) / (1.0 - gwr.influ)
    cv = np.sum(aa**2) / gwr.n
    return cv


def corr(cov):
    invsd = np.diag(1 / np.sqrt(np.diag(cov)))
    cors = np.dot(np.dot(invsd, cov), invsd)
    return cors



"""
GWR is tested against results from GWR4
"""

import os
import numpy as np
from libpysal import io
import libpysal as ps
import unittest
from spglm.family import Gaussian, Poisson, Binomial
from ..sel_bw import Sel_BW
from numpy.testing import assert_allclose


class TestSelBWGaussian(unittest.TestCase):
    def setUp(self):
        data_path = ps.examples.get_path("GData_utm.csv")
        data = io.open(data_path)
        self.coords = list(zip(data.by_col('X'), data.by_col('Y')))
        self.coords_longlat = list(
            zip(data.by_col('Longitud'), data.by_col('Latitude')))
        self.y = np.array(data.by_col('PctBach')).reshape((-1, 1))
        rural = np.array(data.by_col('PctRural')).reshape((-1, 1))
        pov = np.array(data.by_col('PctPov')).reshape((-1, 1))
        black = np.array(data.by_col('PctBlack')).reshape((-1, 1))
        fb = np.array(data.by_col('PctFB')).reshape((-1, 1))
        self.X = np.hstack([rural, pov, black])
        self.mgwr_X = np.hstack([fb, black, rural])

    def test_golden_fixed_AICc(self):
        bw1 = 211020.83
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='bisquare',
                     fixed=True).search(criterion='AICc')
        assert_allclose(bw1, bw2)
        scipy_known = 211025.26298
        scipy = Sel_BW(self.coords, self.y, self.X,
                       kernel='bisquare', fixed=True).search(
                           criterion='AICc', search_method='scipy')
        assert_allclose(scipy_known, scipy, atol=1)

    def test_golden_adapt_AICc(self):
        bw1 = 93.0
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='bisquare',
                     fixed=False).search(criterion='AICc')
        assert_allclose(bw1, bw2)

    def test_golden_adapt_AICc_Longlat(self):
        bw1 = 92.0
        bw2 = Sel_BW(self.coords_longlat, self.y, self.X, kernel='bisquare',
                     fixed=False, spherical=True).search(criterion='AICc')
        assert_allclose(bw1, bw2)

    def test_golden_fixed_AIC(self):
        bw1 = 76201.66
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='gaussian',
                     fixed=True).search(criterion='AIC')
        assert_allclose(bw1, bw2)
        scipy_known = 76199.81
        scipy = Sel_BW(self.coords, self.y, self.X,
                       kernel='gaussian', fixed=True).search(
                           criterion='AIC', search_method='scipy')
        assert_allclose(scipy_known, scipy, atol=1)

    def test_golden_adapt_AIC(self):
        bw1 = 50.0
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='gaussian',
                     fixed=False).search(criterion='AIC')
        assert_allclose(bw1, bw2)

    def test_golden_fixed_BIC(self):
        bw1 = 1117795.47
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='gaussian',
                     fixed=True).search(criterion='BIC')
        assert_allclose(bw1, bw2)
        scipy_known = 1117806.16
        scipy = Sel_BW(self.coords, self.y, self.X,
                       kernel='gaussian', fixed=True).search(
                           criterion='BIC', search_method='scipy')
        assert_allclose(scipy_known, scipy, atol=1)

    def test_golden_adapt_BIC(self):
        bw1 = 62.0
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='gaussian',
                     fixed=False).search(criterion='BIC')
        assert_allclose(bw1, bw2)

    def test_golden_fixed_CV(self):
        bw1 = 130289.26
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='gaussian',
                     fixed=True).search(criterion='CV')
        assert_allclose(bw1, bw2)
        scipy_known = 130363.55
        scipy = Sel_BW(self.coords, self.y, self.X,
                       kernel='gaussian', fixed=True).search(
                           criterion='CV', search_method='scipy')
        assert_allclose(scipy_known, scipy, atol=1)

    def test_golden_adapt_CV(self):
        bw1 = 68.0
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='gaussian',
                     fixed=False).search(criterion='CV')
        assert_allclose(bw1, bw2)

    def test_interval_fixed_AICc(self):
        bw1 = 211025.0
        bw2 = Sel_BW(self.coords, self.y, self.X,
                     kernel='bisquare', fixed=True).search(
                         criterion='AICc', search_method='interval',
                         bw_min=211001.0, bw_max=211035.0, interval=2)
        assert_allclose(bw1, bw2)

    def test_interval_adapt_AICc(self):
        bw1 = 93.0
        bw2 = Sel_BW(self.coords, self.y, self.X,
                     kernel='bisquare', fixed=False).search(
                         criterion='AICc', search_method='interval',
                         bw_min=90.0, bw_max=95.0, interval=1)
        assert_allclose(bw1, bw2)

    def test_interval_fixed_AIC(self):
        bw1 = 76175.0  #76169.00
        bw2 = Sel_BW(self.coords, self.y, self.X,
                     kernel='gaussian', fixed=True).search(
                         criterion='AIC', search_method='interval',
                         bw_min=76161.0, bw_max=76175.0, interval=1)
        assert_allclose(bw1, bw2)

    def test_interval_adapt_AIC(self):
        bw1 = 40.0  #50.0
        bw2 = Sel_BW(self.coords, self.y, self.X,
                     kernel='gaussian', fixed=False).search(
                         criterion='AIC', search_method='interval',
                         bw_min=40.0, bw_max=60.0, interval=2)
        assert_allclose(bw1, bw2)

    def test_interval_fixed_BIC(self):
        bw1 = 279461.0  #279451.00
        bw2 = Sel_BW(self.coords, self.y, self.X,
                     kernel='gaussian', fixed=True).search(
                         criterion='BIC', search_method='interval',
                         bw_min=279441.0, bw_max=279461.0, interval=2)
        assert_allclose(bw1, bw2)

    def test_interval_adapt_BIC(self):
        bw1 = 62.0
        bw2 = Sel_BW(self.coords, self.y, self.X,
                     kernel='gaussian', fixed=False).search(
                         criterion='BIC', search_method='interval',
                         bw_min=52.0, bw_max=72.0, interval=2)
        assert_allclose(bw1, bw2)

    def test_interval_fixed_CV(self):
        bw1 = 130400.0  #130406.00
        bw2 = Sel_BW(self.coords, self.y, self.X,
                     kernel='gaussian', fixed=True).search(
                         criterion='CV', search_method='interval',
                         bw_min=130400.0, bw_max=130410.0, interval=1)
        assert_allclose(bw1, bw2)

    def test_interval_adapt_CV(self):
        bw1 = 62.0  #68.0
        bw2 = Sel_BW(self.coords, self.y, self.X,
                     kernel='gaussian', fixed=False).search(
                         criterion='CV', search_method='interval', bw_min=60.0,
                         bw_max=76.0, interval=2)
        assert_allclose(bw1, bw2)

    def test_MGWR_AICc(self):
        bw1 = [101.0, 101.0, 117.0, 157.0]
        std_y = (self.y - self.y.mean()) / self.y.std()
        std_X = (self.mgwr_X - self.mgwr_X.mean(axis=0)) / self.mgwr_X.std(
            axis=0)
        selector = Sel_BW(self.coords, std_y, std_X, multi=True, constant=True)
        bw2 = selector.search()
        np.testing.assert_allclose(bw1, bw2)

    def test_MGWR_AICc_Longlat(self):
        bw1 = [104.0, 104.0, 103.0, 157.0]
        std_y = (self.y - self.y.mean()) / self.y.std()
        std_X = (self.mgwr_X - self.mgwr_X.mean(axis=0)) / self.mgwr_X.std(
            axis=0)
        selector = Sel_BW(self.coords_longlat, std_y, std_X, multi=True,
                          constant=True, spherical=True)
        bw2 = selector.search()
        assert_allclose(bw1, bw2)


class TestGWRSelBWPoisson(unittest.TestCase):
    def setUp(self):
        data_path = os.path.join(
            os.path.dirname(__file__), 'tokyo/Tokyomortality.csv')
        data = io.open(data_path, mode='Ur')
        self.coords = list(
            zip(data.by_col('X_CENTROID'), data.by_col('Y_CENTROID')))
        self.y = np.array(data.by_col('db2564')).reshape((-1, 1))
        self.off = np.array(data.by_col('eb2564')).reshape((-1, 1))
        OCC = np.array(data.by_col('OCC_TEC')).reshape((-1, 1))
        OWN = np.array(data.by_col('OWNH')).reshape((-1, 1))
        POP = np.array(data.by_col('POP65')).reshape((-1, 1))
        UNEMP = np.array(data.by_col('UNEMP')).reshape((-1, 1))
        self.X = np.hstack([OCC, OWN, POP, UNEMP])

    def test_golden_adapt_AICc_Poisson_w_offset(self):
        bw1 = 95.0
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='bisquare',
                     family=Poisson(), fixed=False,
                     offset=self.off).search(criterion='AICc')
        assert_allclose(bw1, bw2)

    def test_golden_adapt_AICc_Poisson_wo_offset(self):
        bw1 = 51.0
        bw2 = Sel_BW(self.coords, self.y, self.X, kernel='bisquare',
                     family=Poisson(), fixed=False).search(criterion='AICc')
        assert_allclose(bw1, bw2)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from libpysal.common import requires


@requires('matplotlib')
def shift_colormap(cmap, start=0, midpoint=0.5, stop=1.0, name='shiftedcmap'):
    '''
    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and you want the
    middle of the colormap's dynamic range to be at zero

    Parameters
    ----------
    cmap : The matplotlib colormap to be altered
    start : Offset from lowest point in the colormap's range.
      Defaults to 0.0 (no lower ofset). Should be between
      0.0 and `midpoint`.
    midpoint : The new center of the colormap. Defaults to
      0.5 (no shift). Should be between 0.0 and 1.0. In
      general, this should be  1 - vmax/(vmax + abs(vmin))
      For example if your data range from -15.0 to +5.0 and
      you want the center of the colormap at 0.0, `midpoint`
      should be set to  1 - 5/(5 + 15)) or 0.75
    stop : Offset from highets point in the colormap's range.
      Defaults to 1.0 (no upper ofset). Should be between
      `midpoint` and 1.0.
    
    Returns
    -------
    new_cmap : A new colormap that has been shifted. 
    '''

    import matplotlib as mpl
    import matplotlib.pyplot as plt

    cdict = {'red': [], 'green': [], 'blue': [], 'alpha': []}

    # regular index to compute the colors
    reg_index = np.linspace(start, stop, 257)

    # shifted index to match the data
    shift_index = np.hstack([
        np.linspace(0.0, midpoint, 128, endpoint=False),
        np.linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    new_cmap = mpl.colors.LinearSegmentedColormap(name, cdict)
    plt.register_cmap(cmap=new_cmap)

    return new_cmap


@requires('matplotlib')
def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=100):
    '''
    Function to truncate a colormap by selecting a subset of the original colormap's values

    Parameters
    ----------
    cmap : Mmatplotlib colormap to be altered
    minval : Minimum value of the original colormap to include in the truncated colormap
    maxval : Maximum value of the original colormap to include in the truncated colormap
    n : Number of intervals between the min and max values for the gradient of the truncated colormap
          
    Returns
    -------
    new_cmap : A new colormap that has been shifted. 
    '''

    import matplotlib as mpl

    new_cmap = mpl.colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    return new_cmap


@requires('matplotlib')
@requires('geopandas')
def compare_surfaces(data, var1, var2, gwr_t, gwr_bw, mgwr_t, mgwr_bw, name,
                     kwargs1, kwargs2, savefig=None):
    '''
    Function that creates comparative visualization of GWR and MGWR surfaces.

    Parameters
    ----------
    data   : pandas or geopandas Dataframe
             gwr/mgwr results
    var1   : string
             name of gwr parameter estimate column in frame
    var2   : string
             name of mgwr parameter estimate column in frame
    gwr_t  : string
             name of gwr t-values column in frame associated with var1
    gwr_bw : float
             bandwidth for gwr model for var1
    mgwr_t : string
             name of mgwr t-values column in frame associated with var2
    mgwr_bw: float
             bandwidth for mgwr model for var2
    name   : string
             common variable name to use for title
    kwargs1:
             additional plotting arguments for gwr surface
    kwargs2:
             additional plotting arguments for mgwr surface
    savefig: string, optional
             path to save the figure. Default is None. Not to save figure.
    '''
    import matplotlib.pyplot as plt
    import geopandas as gp

    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(45, 20))
    ax0 = axes[0]
    ax0.set_title('GWR ' + name + ' Surface (BW: ' + str(gwr_bw) + ')',
                  fontsize=40)
    ax1 = axes[1]
    ax1.set_title('MGWR ' + name + ' Surface (BW: ' + str(mgwr_bw) + ')',
                  fontsize=40)

    #Set color map
    cmap = plt.cm.seismic

    #Find min and max values of the two combined datasets
    gwr_min = data[var1].min()
    gwr_max = data[var1].max()
    mgwr_min = data[var2].min()
    mgwr_max = data[var2].max()
    vmin = np.min([gwr_min, mgwr_min])
    vmax = np.max([gwr_max, mgwr_max])

    #If all values are negative use the negative half of the colormap
    if (vmin < 0) & (vmax < 0):
        cmap = truncate_colormap(cmap, 0.0, 0.5)
    #If all values are positive use the positive half of the colormap
    elif (vmin > 0) & (vmax > 0):
        cmap = truncate_colormap(cmap, 0.5, 1.0)
    #Otherwise, there are positive and negative values so the colormap so zero is the midpoint
    else:
        cmap = shift_colormap(cmap, start=0.0,
                              midpoint=1 - vmax / (vmax + abs(vmin)), stop=1.)

    #Create scalar mappable for colorbar and stretch colormap across range of data values
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(
        vmin=vmin, vmax=vmax))

    #Plot GWR parameters
    data.plot(var1, cmap=sm.cmap, ax=ax0, vmin=vmin, vmax=vmax, **kwargs1)
    if (gwr_t == 0).any():
        data[gwr_t == 0].plot(color='lightgrey', ax=ax0, **kwargs2)

    #Plot MGWR parameters
    data.plot(var2, cmap=sm.cmap, ax=ax1, vmin=vmin, vmax=vmax, **kwargs1)
    if (mgwr_t == 0).any():
        data[mgwr_t == 0].plot(color='lightgrey', ax=ax1, **kwargs2)

    #Set figure options and plot
    fig.tight_layout()
    fig.subplots_adjust(right=0.9)
    cax = fig.add_axes([0.92, 0.14, 0.03, 0.75])
    sm._A = []
    cbar = fig.colorbar(sm, cax=cax)
    cbar.ax.tick_params(labelsize=50)
    ax0.get_xaxis().set_visible(False)
    ax0.get_yaxis().set_visible(False)
    ax1.get_xaxis().set_visible(False)
    ax1.get_yaxis().set_visible(False)
    if savefig is not None:
        plt.savefig(savefig)
    plt.show()



# GWR Bandwidth selection class

#x_glob parameter does not yet do anything; it is for semiparametric

__author__ = "Taylor Oshan Tayoshan@gmail.com"

import spreg.user_output as USER
import numpy as np
import multiprocessing as mp
from scipy.spatial.distance import pdist
from scipy.optimize import minimize_scalar
from spglm.family import Gaussian, Poisson, Binomial
from .kernels import Kernel
from .gwr import GWR
from .search import golden_section, equal_interval, multi_bw
from .diagnostics import get_AICc, get_AIC, get_BIC, get_CV

getDiag = {'AICc': get_AICc, 'AIC': get_AIC, 'BIC': get_BIC, 'CV': get_CV}


class Sel_BW(object):
    """
    Select bandwidth for kernel

    Methods: p211 - p213, bandwidth selection

    :cite:`fotheringham_geographically_2002`: Fotheringham, A. S., Brunsdon, C., & Charlton, M. (2002).
    Geographically weighted regression: the analysis of spatially varying relationships.

    Parameters
    ----------
    y              : array
                     n*1, dependent variable.
    X_glob         : array
                     n*k1, fixed independent variable.
    X_loc          : array
                     n*k2, local independent variable, including constant.
    coords         : list of tuples
                     (x,y) of points used in bandwidth selection
    family         : family object/instance, optional
                     underlying probability model: Gaussian(), Poisson(),
                     Binomial(). Default is Gaussian().
    offset         : array
                     n*1, the offset variable at the ith location. For Poisson model
                     this term is often the size of the population at risk or
                     the expected size of the outcome in spatial epidemiology
                     Default is None where Ni becomes 1.0 for all locations
    kernel         : string, optional
                     kernel function: 'gaussian', 'bisquare', 'exponential'.
                     Default is 'bisquare'.
    fixed          : boolean
                     True for fixed bandwidth and False for adaptive (NN)
    multi          : True for multiple (covaraite-specific) bandwidths
                     False for a traditional (same for  all covariates)
                     bandwdith; defualt is False.
    constant       : boolean
                     True to include intercept (default) in model and False to exclude
                     intercept.
    spherical      : boolean
                     True for shperical coordinates (long-lat),
                     False for projected coordinates (defalut).

    Attributes
    ----------
    y              : array
                     n*1, dependent variable.
    X_glob         : array
                     n*k1, fixed independent variable.
    X_loc          : array
                     n*k2, local independent variable, including constant.
    coords         : list of tuples
                     (x,y) of points used in bandwidth selection
    family         : string
                     GWR model type: 'Gaussian', 'logistic, 'Poisson''
    kernel         : string
                     type of kernel used and wether fixed or adaptive
    fixed          : boolean
                     True for fixed bandwidth and False for adaptive (NN)
    criterion      : string
                     bw selection criterion: 'AICc', 'AIC', 'BIC', 'CV'
    search_method  : string
                     bw search method: 'golden', 'interval'
    bw_min         : float
                     min value used in bandwidth search
    bw_max         : float
                     max value used in bandwidth search
    interval       : float
                     interval increment used in interval search
    tol            : float
                     tolerance used to determine convergence
    max_iter       : integer
                     max interations if no convergence to tol
    multi          : True for multiple (covaraite-specific) bandwidths
                     False for a traditional (same for  all covariates)
                     bandwdith; defualt is False.
    constant       : boolean
                     True to include intercept (default) in model and False to exclude
                     intercept.
    offset         : array
                     n*1, the offset variable at the ith location. For Poisson model
                     this term is often the size of the population at risk or
                     the expected size of the outcome in spatial epidemiology
                     Default is None where Ni becomes 1.0 for all locations
    spherical      : boolean
                     True for shperical coordinates (long-lat),
                     False for projected coordinates (defalut).
    search_params  : dict
                     stores search arguments
    int_score      : boolan
                     True if adaptive bandwidth is being used and bandwdith
                     selection should be discrete. False
                     if fixed bandwidth is being used and bandwidth does not have
                     to be discrete.
    bw             : scalar or array-like
                     Derived optimal bandwidth(s). Will be a scalar for GWR
                     (multi=False) and a list of scalars for MGWR (multi=True)
                     with one bandwidth for each covariate.
    S              : array
                     n*n, hat matrix derived from the iterative backfitting
                     algorthim for MGWR during bandwidth selection
    R              : array
                     n*n*k, partial hat matrices derived from the iterative
                     backfitting algoruthm for MGWR during bandwidth selection.
                     There is one n*n matrix for each of the k covariates.
    params         : array
                     n*k, calibrated parameter estimates for MGWR based on the
                     iterative backfitting algorithm - computed and saved here to
                     avoid having to do it again in the MGWR object.

    Examples
    --------

    >>> import libpysal as ps
    >>> from mgwr.sel_bw import Sel_BW
    >>> data = ps.io.open(ps.examples.get_path('GData_utm.csv'))
    >>> coords = list(zip(data.by_col('X'), data.by_col('Y')))
    >>> y = np.array(data.by_col('PctBach')).reshape((-1,1))
    >>> rural = np.array(data.by_col('PctRural')).reshape((-1,1))
    >>> pov = np.array(data.by_col('PctPov')).reshape((-1,1))
    >>> african_amer = np.array(data.by_col('PctBlack')).reshape((-1,1))
    >>> X = np.hstack([rural, pov, african_amer])
    
    Golden section search AICc - adaptive bisquare

    >>> bw = Sel_BW(coords, y, X).search(criterion='AICc')
    >>> print(bw)
    93.0

    Golden section search AIC - adaptive Gaussian

    >>> bw = Sel_BW(coords, y, X, kernel='gaussian').search(criterion='AIC')
    >>> print(bw)
    50.0

    Golden section search BIC - adaptive Gaussian

    >>> bw = Sel_BW(coords, y, X, kernel='gaussian').search(criterion='BIC')
    >>> print(bw)
    62.0

    Golden section search CV - adaptive Gaussian

    >>> bw = Sel_BW(coords, y, X, kernel='gaussian').search(criterion='CV')
    >>> print(bw)
    68.0

    Interval AICc - fixed bisquare

    >>> sel = Sel_BW(coords, y, X, fixed=True)
    >>> bw = sel.search(search_method='interval', bw_min=211001.0, bw_max=211035.0, interval=2)
    >>> print(bw)
    211025.0

    """

    def __init__(self, coords, y, X_loc, X_glob=None, family=Gaussian(),
                 offset=None, kernel='bisquare', fixed=False, multi=False,
                 constant=True, spherical=False):
        self.coords = np.array(coords)
        self.y = y
        self.X_loc = X_loc
        if X_glob is not None:
            self.X_glob = X_glob
        else:
            self.X_glob = []
        self.family = family
        self.fixed = fixed
        self.kernel = kernel
        if offset is None:
            self.offset = np.ones((len(y), 1))
        else:
            self.offset = offset * 1.0
        self.multi = multi
        self._functions = []
        self.constant = constant
        self.spherical = spherical
        self.search_params = {}

    def search(self, search_method='golden_section', criterion='AICc',
               bw_min=None, bw_max=None, interval=0.0, tol=1.0e-6,
               max_iter=200, init_multi=None, tol_multi=1.0e-5,
               rss_score=False, max_iter_multi=200, multi_bw_min=[None],
               multi_bw_max=[None
                             ], bws_same_times=3, pool=None, verbose=False):
        """
        Method to select one unique bandwidth for a gwr model or a
        bandwidth vector for a mgwr model.

        Parameters
        ----------
        criterion      : string
                         bw selection criterion: 'AICc', 'AIC', 'BIC', 'CV'
        search_method  : string
                         bw search method: 'golden', 'interval'
        bw_min         : float
                         min value used in bandwidth search
        bw_max         : float
                         max value used in bandwidth search
        multi_bw_min   : list 
                         min values used for each covariate in mgwr bandwidth search.
                         Must be either a single value or have one value for
                         each covariate including the intercept
        multi_bw_max   : list
                         max values used for each covariate in mgwr bandwidth
                         search. Must be either a single value or have one value
                         for each covariate including the intercept
        interval       : float
                         interval increment used in interval search
        tol            : float
                         tolerance used to determine convergence
        max_iter       : integer
                         max iterations if no convergence to tol
        init_multi     : float
                         None (default) to initialize MGWR with a bandwidth
                         derived from GWR. Otherwise this option will choose the
                         bandwidth to initialize MGWR with.
        tol_multi      : convergence tolerence for the multiple bandwidth
                         backfitting algorithm; a larger tolerance may stop the
                         algorith faster though it may result in a less optimal
                         model
        max_iter_multi : max iterations if no convergence to tol for multiple
                         bandwidth backfitting algorithm
        rss_score      : True to use the residual sum of sqaures to evaluate
                         each iteration of the multiple bandwidth backfitting
                         routine and False to use a smooth function; default is
                         False
        bws_same_times : If bandwidths keep the same between iterations for
                         bws_same_times (default 3) in backfitting, then use the
                         current set of bandwidths as final bandwidths.
        pool           : A multiprocessing Pool object to enbale parallel fitting;
                         default is None
        verbose        : Boolean
                         If true, bandwidth searching history is printed out; default is False.

        Returns
        -------
        bw             : scalar or array
                         optimal bandwidth value or values; returns scalar for
                         multi=False and array for multi=True; ordering of bandwidths
                         matches the ordering of the covariates (columns) of the
                         designs matrix, X
        """
        k = self.X_loc.shape[1]
        if self.constant:  #k is the number of covariates
            k += 1
        self.search_method = search_method
        self.criterion = criterion
        self.bw_min = bw_min
        self.bw_max = bw_max
        self.bws_same_times = bws_same_times
        self.pool = pool
        self.verbose = verbose

        if len(multi_bw_min) == k:
            self.multi_bw_min = multi_bw_min
        elif len(multi_bw_min) == 1:
            self.multi_bw_min = multi_bw_min * k
        else:
            raise AttributeError(
                "multi_bw_min must be either a list containing"
                " a single entry or a list containing an entry for each of k"
                " covariates including the intercept")

        if len(multi_bw_max) == k:
            self.multi_bw_max = multi_bw_max
        elif len(multi_bw_max) == 1:
            self.multi_bw_max = multi_bw_max * k
        else:
            raise AttributeError(
                "multi_bw_max must be either a list containing"
                " a single entry or a list containing an entry for each of k"
                " covariates including the intercept")

        self.interval = interval
        self.tol = tol
        self.max_iter = max_iter
        self.init_multi = init_multi
        self.tol_multi = tol_multi
        self.rss_score = rss_score
        self.max_iter_multi = max_iter_multi
        self.search_params['search_method'] = search_method
        self.search_params['criterion'] = criterion
        self.search_params['bw_min'] = bw_min
        self.search_params['bw_max'] = bw_max
        self.search_params['interval'] = interval
        self.search_params['tol'] = tol
        self.search_params['max_iter'] = max_iter
        #self._check_min_max()

        self.int_score = not self.fixed

        if self.multi:
            self._mbw()
            self.params = self.bw[3]  #params n by k
            self.bw_init = self.bw[
                -1]  #scalar, optimal bw from initial gwr model
        else:
            self._bw()

        self.pool = None
        return self.bw[0]

    def _bw(self):
        gwr_func = lambda bw: getDiag[self.criterion](GWR(
            self.coords, self.y, self.X_loc, bw, family=self.family, kernel=
            self.kernel, fixed=self.fixed, constant=self.constant, offset=self.
            offset, spherical=self.spherical).fit(lite=True, pool=self.pool))

        self._optimized_function = gwr_func

        if self.search_method == 'golden_section':
            a, c = self._init_section(self.X_glob, self.X_loc, self.coords,
                                      self.constant)
            delta = 0.38197  #1 - (np.sqrt(5.0)-1.0)/2.0
            self.bw = golden_section(a, c, delta, gwr_func, self.tol,
                                     self.max_iter, self.int_score,
                                     self.verbose)
        elif self.search_method == 'interval':
            self.bw = equal_interval(self.bw_min, self.bw_max, self.interval,
                                     gwr_func, self.int_score, self.verbose)
        elif self.search_method == 'scipy':
            self.bw_min, self.bw_max = self._init_section(
                self.X_glob, self.X_loc, self.coords, self.constant)
            if self.bw_min == self.bw_max:
                raise Exception(
                    'Maximum bandwidth and minimum bandwidth must be distinct for scipy optimizer.'
                )
            self._optimize_result = minimize_scalar(
                gwr_func, bounds=(self.bw_min, self.bw_max), method='bounded')
            self.bw = [self._optimize_result.x, self._optimize_result.fun, []]
        else:
            raise TypeError('Unsupported computational search method ',
                            self.search_method)

    def _mbw(self):
        y = self.y
        if self.constant:
            X = USER.check_constant(self.X_loc)
        else:
            X = self.X_loc
        n, k = X.shape
        family = self.family
        offset = self.offset
        kernel = self.kernel
        fixed = self.fixed
        spherical = self.spherical
        coords = self.coords
        search_method = self.search_method
        criterion = self.criterion
        bw_min = self.bw_min
        bw_max = self.bw_max
        multi_bw_min = self.multi_bw_min
        multi_bw_max = self.multi_bw_max
        interval = self.interval
        tol = self.tol
        max_iter = self.max_iter
        bws_same_times = self.bws_same_times

        def gwr_func(y, X, bw):
            return GWR(coords, y, X, bw, family=family, kernel=kernel,
                       fixed=fixed, offset=offset, constant=False,
                       spherical=self.spherical, hat_matrix=False).fit(
                           lite=True, pool=self.pool)

        def bw_func(y, X):
            selector = Sel_BW(coords, y, X, X_glob=[], family=family,
                              kernel=kernel, fixed=fixed, offset=offset,
                              constant=False, spherical=self.spherical)
            return selector

        def sel_func(bw_func, bw_min=None, bw_max=None):
            return bw_func.search(
                search_method=search_method, criterion=criterion,
                bw_min=bw_min, bw_max=bw_max, interval=interval, tol=tol,
                max_iter=max_iter, pool=self.pool, verbose=False)

        self.bw = multi_bw(self.init_multi, y, X, n, k, family, self.tol_multi,
                           self.max_iter_multi, self.rss_score, gwr_func,
                           bw_func, sel_func, multi_bw_min, multi_bw_max,
                           bws_same_times, verbose=self.verbose)

    def _init_section(self, X_glob, X_loc, coords, constant):
        if len(X_glob) > 0:
            n_glob = X_glob.shape[1]
        else:
            n_glob = 0
        if len(X_loc) > 0:
            n_loc = X_loc.shape[1]
        else:
            n_loc = 0
        if constant:
            n_vars = n_glob + n_loc + 1
        else:
            n_vars = n_glob + n_loc
        n = np.array(coords).shape[0]

        if self.int_score:
            a = 40 + 2 * n_vars
            c = n
        else:
            sq_dists = pdist(coords)
            a = np.min(sq_dists) / 2.0
            c = np.max(sq_dists) * 2.0

        if self.bw_min is not None:
            a = self.bw_min
        if self.bw_max is not None:
            c = self.bw_max

        return a, c
